import dataclasses
from typing import Optional
from xmlrpc.client import Boolean
from prettytable import MARKDOWN, PrettyTable

from wton.tonclient.utils import KeyStores, Whitelist, KeyStore
from wton.tonclient import TonClient, DAppTonClient
from wton.config import init_config, Config, TonProviderEnum
from wton.utils import storage


@dataclasses.dataclass
class SharedObject:
    config: Config
    ton_client: TonClient
    specific_config_path: Optional[str]
    keystores: Optional[KeyStores] = None
    keystore: Optional[KeyStore] = None
    keystore_password: Optional[str] = None
    whitelist: Optional[Whitelist] = None
    debug_mode: Boolean = False


def init_shared_object(specific_config_path: str = None) -> SharedObject:
    config = init_config(
        specific_config_path) if specific_config_path is not None else init_config()
    ton_client = __get_ton_client(config)

    return SharedObject(
        config=config, specific_config_path=specific_config_path, ton_client=ton_client)


def setup_app(config: Config):
    for default_dir_path in [config.wton.workdir,
                             config.wton.keystores_path]:
        storage.ensure_dir_exists(default_dir_path)


def md_table() -> PrettyTable:
    table = PrettyTable()
    table.set_style(MARKDOWN)
    return table

def __get_ton_client(config: Config):
    if config.wton.provider == TonProviderEnum.dapp:
        return DAppTonClient(config)
    else:
        raise NotImplementedError



def new_keystore_password_is_valid(password: str):
    if len(password) < 6:
        return False

    return True



from json import JSONDecodeError
import sys
from pydantic import ValidationError

from wton.config import ConfigNotFoundError
from .._utils import init_shared_object, setup_app
from ._utils import echo_error
from ._sets import EntrypointSet


def main():
    try:
        context = init_shared_object()
        setup_app(context.config)

    except (FileNotFoundError, JSONDecodeError, ConfigNotFoundError, ValidationError) as e:
        echo_error(e)

    if len(sys.argv) == 2 and sys.argv[1] == "--debug":
        context.debug_mode = True

    try:
        EntrypointSet(context).show()
    except KeyboardInterrupt:
        pass



from ._utils import concat_bytes, move_to_end, tree_walk, crc32c, \
    crc16, read_n_bytes_uint_from_array, compare_bytes, sign_message, b64str_to_bytes, \
    b64str_to_hex, bytes_to_b64str
from ._address import Address
from ._currency import to_nano, from_nano, TonCurrencyEnum
from ._exceptions import InvalidAddressError

__all__ = [
    'Address',
    'InvalidAddressError',

    'concat_bytes',
    'move_to_end',
    'tree_walk',
    'crc32c',
    'crc16',
    'read_n_bytes_uint_from_array',
    'compare_bytes',
    'sign_message',
    'b64str_to_bytes',
    'b64str_to_hex',
    'bytes_to_b64str',

    'to_nano',
    'from_nano',
    'TonCurrencyEnum',
]



from typing import Any, List, Tuple, Union
from wton import settings
from wton.utils import storage
from ._config import ConfigLocation, Config, TonNetworkEnum
from ._exceptions import ConfigNotFoundError
from ._wton import WtonConfig, TonProviderEnum
from ._provider import ProviderConfig


def init_config(specific_config_path: str = None) -> Config:
    config = Config()

    for config_path in settings.CONFIG_PATHS:
        config_to_merge = __get_config(config_path)
        if config_to_merge is not None:
            config = __merge_configs(config_to_merge, config)

    if specific_config_path is not None:
        config_to_merge = __get_config(specific_config_path, raise_error=True)
        config = __merge_configs(config_to_merge, config)

    return config


def unset_config_field(config_location: Union[str, ConfigLocation], name: str):
    config_path = __get_config_path_from_location(config_location)
    config = __get_config(config_path)

    if name not in Config.field_names() or config is None:
        return

    config_dict = config.to_nondefault_dict_without_field(name)
    storage.save_yaml(config_path, config_dict)


def update_config_field(config_location: Union[str, ConfigLocation], name: str, value: Any):
    config_path = __get_config_path_from_location(config_location)
    config = __get_config(config_path)
    if config is None:
        config = Config()

    config.update_value(name, value)

    storage.save_yaml(config_path, config.dict(exclude_unset=True))


def get_configs_with_origins(specific_config_path: str) -> Tuple[List[Config], List[str]]:
    configs = []
    origins = []
    config_paths = settings.CONFIG_PATHS

    if specific_config_path:
        config_paths.append(specific_config_path)

    for config_path in config_paths:
        config = __get_config(config_path)
        if config is not None:
            configs.append(config)
            origins.append(config_path)

    return configs, origins


def get_config(config_location: Union[str, ConfigLocation]) -> Union[Config, None]:
    config_path = __get_config_path_from_location(config_location)

    return __get_config(config_path)


def __get_config(config_path: str, raise_error: bool = False) -> Union[Config, None]:
    if not storage.exists(config_path):
        if raise_error:
            raise ConfigNotFoundError(
                f"Incorrect config file path: {config_path}")
        return None

    new_config_dict = storage.read_yaml(config_path)
    if new_config_dict is None:
        return None

    return Config.parse_obj(new_config_dict)


def __merge_dicts(dict1, dict2):
    for k in set(dict1.keys()).union(dict2.keys()):
        if k in dict1 and k in dict2:
            if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                yield (k, dict(__merge_dicts(dict1[k], dict2[k])))
            else:
                yield (k, dict2[k])
        elif k in dict1:
            yield (k, dict1[k])
        else:
            yield (k, dict2[k])


def __merge_configs(merge_to: Config, merge_from: Config) -> Config:
    merged = __merge_dicts(merge_from.dict(exclude_unset=True),
                           merge_to.dict(exclude_unset=True))

    return Config.parse_obj(merged)


def __get_config_path_from_location(config_location: Union[str, ConfigLocation]) -> str:
    if config_location == ConfigLocation.global_location:
        return settings.GLOBAL_CONFIG_PATH
    elif config_location == ConfigLocation.local_location:
        return settings.LOCAL_CONFIG_PATH
    else:
        return config_location



from collections import OrderedDict
import inquirer

from wton.tonclient.utils import Whitelist
from ._base import BaseSet
from ..._utils import SharedObject, md_table
from .._utils import echo_success
from wton.tonsdk.utils import Address


class WhitelistSet(BaseSet):
    def __init__(self, ctx: SharedObject) -> None:
        super().__init__(ctx)
        ctx.whitelist = Whitelist(
            ctx.config.wton.whitelist_path)

    def _handlers(self) -> OrderedDict:
        ord_dict = OrderedDict()
        ord_dict["List contacts"] = self._handle_list
        ord_dict["Add contact"] = self._handle_add
        ord_dict["Get contact"] = self._handle_get
        ord_dict["Edit contact"] = self._handle_edit
        ord_dict["Delete contact"] = self._handle_delete
        ord_dict["Back"] = self._handle_exit
        return ord_dict

    def _handle_list(self):
        questions = [
            inquirer.Confirm(
                "verbose", message='Show balances?', default=True),
        ]
        verbose = self._prompt(questions)["verbose"]

        with self._processing():
            table = md_table()
            field_names = ["Name", "Address"]
            if verbose:
                field_names += ['State', 'Balance']
            table.field_names = field_names

            contacts = self.ctx.whitelist.contacts
            if verbose:
                contact_infos = self.ctx.ton_client.get_addresses_information(
                    [contact.address for contact in contacts])

                for contact, contact_info in zip(contacts, contact_infos):
                    table.add_row([contact.name, contact.address,
                                   contact_info.state, contact_info.balance])

            else:
                for contact in self.ctx.whitelist.contacts:
                    table.add_row([contact.name, contact.address])

        echo_success(table, only_msg=True)

    def _handle_add(self):
        questions = [
            inquirer.Text(
                "name", message='Enter the name'),
            inquirer.Text(
                "address", message='Enter the address'),
        ]
        ans = self._prompt(questions)
        name = ans["name"]
        address = ans["address"]
        self.ctx.whitelist.add_contact(name, address, save=True)

        echo_success()

    def _handle_get(self):
        name = self.__select_contact_or_false('Contact to edit')
        if name == False:
            return

        contact = self.ctx.whitelist.get_contact(name, raise_none=True)

        addr = Address(contact.address)
        echo_success(
            f"Raw address: {addr.to_string(False, False, False)}", True)
        echo_success(
            f"Nonbounceable address: {addr.to_string(True, True, False)}", True)
        echo_success(
            f"Bounceable address: {addr.to_string(True, True, True)}", True)

    def _handle_edit(self):
        name = self.__select_contact_or_false('Contact to edit')
        if name == False:
            return

        questions = [
            inquirer.Text(
                "name", message='Enter new name (enter to skip)'),
            inquirer.Text(
                "address", message='Enter the address (enter to skip)'),
        ]
        ans = self._prompt(questions)
        new_name = ans["name"]
        new_address = ans["address"]

        self.ctx.whitelist.edit_contact(
            name, new_name, new_address, save=True)

        echo_success()

    def _handle_delete(self):
        name = self.__select_contact_or_false('Contact to edit')
        if name == False:
            return

        self.ctx.whitelist.delete_contact(name, save=True)

        echo_success()

    def __select_contact_or_false(self, message):
        if self.ctx.whitelist.contacts:
            questions = [
                inquirer.List(
                    "contact",
                    message=message,
                    choices=[
                        record.name for record in self.ctx.whitelist.contacts],
                    carousel=True
                )
            ]
            return self._prompt(questions)["contact"]

        echo_success("You do not have any contacts yet.")
        return False



import click

from ._base_cmd import cli, SharedObject
from .._utils import CustomClickException, with_keystores, with_keystore, click_echo_success
from ..._utils import new_keystore_password_is_valid
from wton.tonclient.utils import KeyStoreAlreadyExistsError, KeyStoreInvalidPasswordError, InvalidMnemonicsError
from wton import settings


@cli.group()
def keystore():
    """
    Operate with keystores
    """


def __validate_password(ctx, param, value):
    if not new_keystore_password_is_valid(value):
        raise CustomClickException(
            'password need to be at least 6 characters long')
    return value


@keystore.command()
@with_keystores
@click.argument('name', required=True)
@click.password_option(default=settings.KEYSTORE_PASSWORD,
                       confirmation_prompt=False, prompt=False if settings.KEYSTORE_PASSWORD else True,
                       callback=__validate_password)
@click.pass_obj
def new(shared_object: SharedObject, name, password):
    """
    Create new .keystore file with a given name.
    """
    try:
        shared_object.keystores.create_new_keystore(name, password, save=True)
    except KeyStoreAlreadyExistsError as e:
        raise CustomClickException(e)

    click_echo_success(f"keystore {name} has been created. To use it run 'wton config wton.keystore_name {name}'.")

@keystore.command()
@with_keystores
@click.pass_obj
def list(shared_object: SharedObject):
    """
    List all .keystore files in a current keystore workdir.
    """
    for keystore in shared_object.keystores.keystore_paths.keys():
        click.echo(keystore)


@keystore.command()
@with_keystore(password_required=True)
@click.argument('backup_file_path', required=True)
@click.option('--yes', '-y', 'is_sure', is_flag=True, help='Do not show the prompt.')
@click.pass_obj
def backup(shared_object: SharedObject, backup_file_path: str, is_sure: bool):
    """
    Backup the keystore into a specified file.
    """
    if not is_sure:
        click.confirm(
            f'Backup stores keys in UNENCRYPTED FORM. Are you sure want to export unencrypted keys to disk?', abort=True)
    try:
        shared_object.keystore.backup(
            backup_file_path, shared_object.keystore_password)
    except (KeyStoreInvalidPasswordError) as e:
        raise CustomClickException(e)

    click_echo_success(f"keystore {shared_object.config.wton.keystore_name} has been backed up to the {backup_file_path}.")


@keystore.command()
@with_keystores
@click.argument('name', required=True)
@click.argument('backup_file_path', required=True)
@click.password_option(default=settings.KEYSTORE_PASSWORD,
                       confirmation_prompt=False, prompt=False if settings.KEYSTORE_PASSWORD else True,
                       callback=__validate_password)
@click.option('--from-ton-cli', 'from_ton_cli', is_flag=True, help='Restore from the ton-cli util.')
@click.pass_obj
def restore(shared_object: SharedObject, name: str, backup_file_path: str, password: str, from_ton_cli: bool):
    """
    Restore the keystore from a specified file.
    """
    try:
        if from_ton_cli:
            shared_object.keystores.restore_ton_cli_keystore(
                name, backup_file_path, password)
        else:
            shared_object.keystores.restore_wton_keystore(
                name, backup_file_path, password)
    except (InvalidMnemonicsError, KeyStoreAlreadyExistsError) as e:
        raise CustomClickException(e)

    click_echo_success(f"keystore {name} has been restored.")



import asyncio
from typing import List
from datetime import datetime
from gql.transport.exceptions import TransportQueryError

from wton.tonsdk.utils import to_nano, TonCurrencyEnum, from_nano, Address, b64str_to_bytes, bytes_to_b64str
from wton.tonsdk.boc import Cell
from wton.tonsdk.contract.wallet import SendModeEnum, WalletContract
from wton.config import Config
from wton.tonsdk.provider.dapp import DAppClient, DAppWrongResult, BroadcastQuery
from ..._exceptions import TonDappError
from .._base import TonClient, AddressInfoResult, BroadcastResult, BroadcastStatusEnum
from ._queries import DAppQueries


class DAppTonClient(TonClient):
    def __init__(self, config: Config):
        self.config = config
        self.loop = asyncio.get_event_loop()
        self._provider: DAppClient = None

    @property
    def provider(self):
        if self._provider is None:
            self._provider = DAppClient(graphql_url=self.config.provider.dapp.graphql_url,
                                        broadcast_url=self.config.provider.dapp.broadcast_url,
                                        api_key=self.config.provider.dapp.api_key)

        return self._provider

    def get_address_information(self, address: str,
                                currency_to_show: TonCurrencyEnum = TonCurrencyEnum.ton):
        return self.get_addresses_information([address], currency_to_show)[0]

    def get_addresses_information(self, addresses: List[str],
                                  currency_to_show: TonCurrencyEnum = TonCurrencyEnum.ton):
        if not addresses:
            return []

        address_ids = [Address(addr).to_string(
            False, False, False) for addr in addresses]
        query = DAppQueries.accounts(address_ids)
        results = self._run(self.provider.query(
            [query]), single_query=True)['accounts']

        address_infos = [None] * len(address_ids)
        for result in results:
            idx = address_ids.index(result['id'])
            address_infos[idx] = self._parse_addr_info(result)
        for i in range(len(address_infos)):
            if address_infos[i] is None:
                address_infos[i] = AddressInfoResult(
                    address=addresses[i], state="Uninit", balance="0")

        return address_infos

    def seqno(self, addr: str):
        if self.get_address_information(addr).seqno:
            return int(self.get_address_information(addr).seqno)

        return 0

    def deploy_wallet(self, wallet: WalletContract, wait_for_result=False):
        timeout = 30 if wait_for_result else 0
        query = wallet.create_init_external_message()
        base64_boc = bytes_to_b64str(query["message"].to_boc(False))
        result = self._run(self.provider.broadcast(
            [BroadcastQuery(boc=base64_boc, timeout=timeout)]))

        return self._parse_broadcast_result(result, wait_for_result)

    def transfer(self, from_wallet: WalletContract, to_addr: str, amount: TonCurrencyEnum.ton, payload=None,
                 send_mode=SendModeEnum.ignore_errors | SendModeEnum.pay_gas_separately, wait_for_result=False):
        timeout = 30 if wait_for_result else 0
        seqno = self.seqno(from_wallet.address.to_string())
        query = from_wallet.create_transfer_message(
            to_addr=Address(to_addr).to_string(False, False, True),  # TODO: set is bounceable or not
            amount=to_nano(amount, TonCurrencyEnum.ton),
            seqno=seqno,
            payload=payload,
            send_mode=send_mode,
        )
        msg_boc = query["message"].to_boc(False)
        base64_boc = bytes_to_b64str(msg_boc)
        result = self._run(self.provider.broadcast(
            [BroadcastQuery(boc=base64_boc, timeout=timeout)]))

        return self._parse_broadcast_result(result, wait_for_result)

    def _run(self, to_run, *, single_query=True):
        try:
            results = self.loop.run_until_complete(to_run)
        except TransportQueryError as e:
            raise TonDappError(str(e))
        except DAppWrongResult as e:
            raise TonDappError(str(e))

        if single_query:
            return results[0]

        return results

    def _parse_addr_info(self, result: dict, currency_to_show: TonCurrencyEnum = TonCurrencyEnum.ton):
        return AddressInfoResult(
            address=result['address'],
            # contract_type='wallet v3r2',  # FIXME
            seqno=self._get_seqno(result),
            state=result['acc_type_name'],
            balance=self._get_balance(result, currency_to_show),
            last_activity=self._get_last_paid(result),
            code=result['code'],
            data=result['data'],
        )

    def _get_seqno(self, result):
        if result['acc_type_name'] in ["Active", "Frozen"]:
            # TODO: check contract type and version
            data_cell = Cell.one_from_boc(b64str_to_bytes(result["data"]))
            if len(data_cell.bits) > 32:
                seqno = 0
                for bit in data_cell.bits[:32]:
                    seqno = (seqno << 1) | bit
                return seqno

        return 0

    def _get_balance(self, result: dict, currency_to_show):
        if "balance" in result and result["balance"]:
            if int(result["balance"]) < 0:
                balance = 0
            else:
                balance = from_nano(
                    float(result["balance"]), currency_to_show)
        else:
            balance = 0

        return balance

    def _get_last_paid(self, result: dict):
        if "last_paid" in result and result["last_paid"]:
            return str(datetime.utcfromtimestamp(
                result['last_paid']).strftime('%Y-%m-%d %H:%M:%S'))

    def _parse_broadcast_result(self, result, waited) -> BroadcastResult:
        if "status" in result:
            if result["status"] == 1 and waited:
                status = BroadcastStatusEnum.commited
            else:
                status = BroadcastStatusEnum.broadcasted
        else:
            status = BroadcastStatusEnum.failed

        return BroadcastResult(waited=waited, status=status)



from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel

from wton.config import Config
from wton.tonsdk.utils import TonCurrencyEnum
from wton.tonsdk.contract.wallet import SendModeEnum, WalletContract


class AddressInfoResult(BaseModel):
    address: Optional[str] = None
    contract_type: Optional[str] = None
    seqno: Optional[str] = None
    state: Optional[str] = None
    balance: Optional[str] = None
    last_activity: Optional[str] = None
    code: Optional[str] = None
    data: Optional[str] = None


class BroadcastStatusEnum(str, Enum):
    broadcasted = "broadcasted"
    commited = "commited"
    failed = "failed"


class BroadcastResult(BaseModel):
    waited: bool
    status: BroadcastStatusEnum

    class Config:
        use_enum_values = True
        validate_assignment = True

    def __str__(self):
        if self.status == BroadcastStatusEnum.commited:
            return "Transaction has been commited."

        if self.status == BroadcastStatusEnum.broadcasted:
            if self.waited:
                return "Transaction has been sent but hasn't been committed into blockchain during 30 seconds."
            else:
                return "Transaction has been sent."

        return "Transaction hasn't been sent."


class TonClient(ABC):
    @abstractmethod
    def __init__(self, config: Config):
        raise NotImplementedError

    @abstractmethod
    def get_address_information(self, address: str,
                                currency_to_show: TonCurrencyEnum = TonCurrencyEnum.ton) -> AddressInfoResult:
        raise NotImplementedError

    @abstractmethod
    def get_addresses_information(self, addresses: List[str],
                                  currency_to_show: TonCurrencyEnum = TonCurrencyEnum.ton) -> List[AddressInfoResult]:
        raise NotImplementedError

    @abstractmethod
    def seqno(self, addr: str) -> int:
        raise NotImplementedError

    @abstractmethod
    def deploy_wallet(self, wallet: WalletContract, wait_for_result=False) -> BroadcastResult:
        raise NotImplementedError

    @abstractmethod
    def transfer(self, from_wallet: WalletContract, to_addr: str, amount: TonCurrencyEnum.ton, payload=None,
                 send_mode=SendModeEnum.ignore_errors | SendModeEnum.pay_gas_separately, wait_for_result=False) -> BroadcastResult:
        raise NotImplementedError



from functools import wraps
from click import ClickException
import click
from click._compat import get_text_stderr
from gettext import gettext as _
import typing as t
from json import JSONDecodeError
from wton import settings

from wton.tonclient.utils import Whitelist, KeyStores, \
    KeyStoreDoesNotExistError, InvalidKeyStoreError, KeyStoreIsNotSpecifiedError
from wton.tonclient import TonError


class CustomClickException(ClickException):
    def show(self, file: t.Optional[t.IO] = None) -> None:
        if file is None:
            file = get_text_stderr()

        click.echo(_("{error}: {message}").format(
            error=click.style('Error', fg='red'),
            message=self.format_message()), file=file)


def click_echo_success(msg):
    click.echo("{success}: {message}".format(
        success=click.style('Success', fg='green'),
        message=msg
    ))


def with_whitelist(func):
    @click.pass_obj
    @wraps(func)
    def wrapper(shared_object, *args, **kwargs):
        try:
            shared_object.whitelist = Whitelist(
                shared_object.config.wton.whitelist_path)
        except JSONDecodeError:
            raise CustomClickException(
                f"Invalid json in the whitelist file: {shared_object.config.wton.whitelist_path}")

        return func(*args, **kwargs)

    return wrapper


def with_keystores(func):
    @click.pass_obj
    @wraps(func)
    def wrapper(shared_object, *args, **kwargs):
        shared_object.keystores = KeyStores(
            shared_object.config.wton.keystores_path)

        return func(*args, **kwargs)

    return wrapper


def with_keystore(password_required: bool):
    def without_password(func):
        @click.pass_obj
        @wraps(func)
        def without_password_wrapper(shared_object, *args, **kwargs):
            keystores = KeyStores(
                shared_object.config.wton.keystores_path)

            try:
                shared_object.keystore = keystores.get_keystore(
                    shared_object.config.wton.keystore_name, raise_none=True)
            except (KeyStoreDoesNotExistError, InvalidKeyStoreError, KeyStoreIsNotSpecifiedError) as e:
                raise CustomClickException(e)

            return func(*args, **kwargs)
        return without_password_wrapper

    def with_password(func):
        @click.pass_obj
        @click.password_option(default=settings.KEYSTORE_PASSWORD,
                               confirmation_prompt=False, prompt=False if settings.KEYSTORE_PASSWORD else True)
        @wraps(func)
        def with_password_wrapper(shared_object, password, *args, **kwargs):
            keystores = KeyStores(
                shared_object.config.wton.keystores_path)

            try:
                shared_object.keystore = keystores.get_keystore(
                    shared_object.config.wton.keystore_name, raise_none=True)
                shared_object.keystore_password = password
            except (KeyStoreDoesNotExistError, InvalidKeyStoreError, KeyStoreIsNotSpecifiedError) as e:
                raise CustomClickException(e)

            return func(*args, **kwargs)

        return with_password_wrapper

    return with_password if password_required else without_password


def click_ton_exception_handler(exception: TonError):
    raise CustomClickException(exception)


class HiddenPassword(object):
    def __init__(self, password=''):
        self.password = password

    def __str__(self):
        return '*' * len(self.password)



from hashlib import sha256
import math
import copy

from ._bit_string import BitString
from ..utils import concat_bytes, tree_walk, crc32c, read_n_bytes_uint_from_array, compare_bytes


class Cell:
    REACH_BOC_MAGIC_PREFIX = bytes.fromhex('B5EE9C72')
    LEAN_BOC_MAGIC_PREFIX = bytes.fromhex('68ff65f3')
    LEAN_BOC_MAGIC_PREFIX_CRC = bytes.fromhex('acc3a728')

    def __init__(self):
        self.bits = BitString(1023)
        self.refs = []
        self.is_exotic = False

    def __str__(self) -> str:
        return "BITS: {}\nREFS: {}\nIS_EXOTIC: {}".format(self.bits, self.refs, self.is_exotic)

    def bytes_hash(self):
        return bytes.fromhex(sha256(self.bytes_repr()).hexdigest())

    def bytes_repr(self):
        repr_array = []

        repr_array.append(self.get_data_with_descriptors())

        for r in self.refs:
            repr_array.append(r.get_max_depth_as_array())

        for r in self.refs:
            repr_array.append(r.bytes_hash())

        x = bytes()
        for r in repr_array:
            x = concat_bytes(x, r)

        return x

    def write_cell(self, another_cell):
        self.bits.write_bit_string(another_cell.bits)
        self.refs += another_cell.refs

    def get_data_with_descriptors(self):
        d1 = self.get_refs_descriptor()
        d2 = self.get_bits_descriptor()
        tuBits = self.bits.get_top_upped_array()

        return concat_bytes(concat_bytes(d1, d2), tuBits)

    def get_bits_descriptor(self):
        d2 = bytearray([0])
        d2[0] = math.ceil(self.bits.cursor / 8) + \
            math.floor(self.bits.cursor / 8)
        return d2

    def get_refs_descriptor(self):
        d1 = bytearray([0])
        d1[0] = len(self.refs) + self.is_exotic * 8 + self.get_max_level() * 32
        return d1

    def get_max_level(self):
        # FIXME: looks strange! nothing happens here. Compare with get_max_depth.
        max_level = 0
        for r in self.refs:
            r_max_level = r.get_max_level()
            if r_max_level > max_level:
                max_level = r_max_level
        return max_level

    def get_max_depth_as_array(self):
        max_depth = self.get_max_depth()
        d = bytearray([0, 0])
        d[1] = max_depth % 256
        d[0] = math.floor(max_depth / 256)
        return d

    def get_max_depth(self):
        max_depth = 0
        if len(self.refs) > 0:
            for r in self.refs:
                r_max_depth = r.get_max_depth()
                if r_max_depth > max_depth:
                    max_depth = r_max_depth
            max_depth += 1
        return max_depth

    def tree_walk(self):
        return tree_walk(self, [], {})

    def is_explicitly_stored_hashes(self):
        return 0

    def serialize_for_boc(self, cells_index, ref_size):
        repr_arr = []

        repr_arr.append(self.get_data_with_descriptors())
        if self.is_explicitly_stored_hashes():
            raise Exception("Cell hashes explicit storing is not implemented")

        for ref in self.refs:
            ref_hash = ref.bytes_hash()
            ref_index_int = cells_index[ref_hash]
            ref_index_hex = format(ref_index_int, 'x')
            if len(ref_index_hex) % 2:
                ref_index_hex = '0' + ref_index_hex
            reference = bytes.fromhex(ref_index_hex)
            repr_arr.append(reference)

        x = b''
        for data in repr_arr:
            x = concat_bytes(x, data)

        return x

    def boc_serialization_size(self, cells_index, ref_size):
        return len(self.serialize_for_boc(cells_index, ref_size))

    def to_boc(self, has_idx=True, hash_crc32=True, has_cache_bits=False, flags=0):
        root_cell = copy.deepcopy(self)

        all_cells = root_cell.tree_walk()
        topological_order = all_cells[0]
        cells_index = all_cells[1]

        cells_num = len(topological_order)
        # Minimal number of bits to represent reference (unused?)
        s = len("{0:b}".format(cells_num))
        s_bytes = min(math.ceil(s / 8), 1)
        full_size = 0
        size_index = []
        for cell_info in topological_order:
            size_index.append(full_size)
            full_size += cell_info[1].boc_serialization_size(
                cells_index, s_bytes)

        offset_bits = len("{0:b}".format(full_size))
        offset_bytes = max(math.ceil(offset_bits / 8), 1)

        serialization = BitString(
            (1023 + 32 * 4 + 32 * 3) * len(topological_order))
        serialization.write_bytes(Cell.REACH_BOC_MAGIC_PREFIX)
        settings = bytes(''.join(['1' if i else '0' for i in [
                         has_idx, hash_crc32, has_cache_bits]]), 'utf-8')
        serialization.write_bit_array(settings)
        serialization.write_uint(flags, 2)
        serialization.write_uint(s_bytes, 3)
        serialization.write_uint8(offset_bytes)
        serialization.write_uint(cells_num, s_bytes * 8)
        serialization.write_uint(1, s_bytes * 8)  # One root for now
        serialization.write_uint(0, s_bytes * 8)  # Complete BOCs only
        serialization.write_uint(full_size, offset_bytes * 8)
        serialization.write_uint(0, s_bytes * 8)  # Root shoulh have index 0

        if has_idx:
            for [cell_data, index] in topological_order:  # ?
                serialization.write_uint(size_index[index], offset_bytes * 8)

        for cell_info in topological_order:
            ref_cell_ser = cell_info[1].serialize_for_boc(cells_index, s_bytes)
            serialization.write_bytes(ref_cell_ser)

        ser_arr = serialization.get_top_upped_array()
        if hash_crc32:
            ser_arr += crc32c(ser_arr)

        return ser_arr

    @staticmethod
    def one_from_boc(serialized_boc):
        cells = deserialize_boc(serialized_boc)

        if len(cells) != 1:
            raise Exception("Expected 1 root cell")

        return cells[0]


def deserialize_cell_data(cell_data, reference_index_size):
    if len(cell_data) < 2:
        raise Exception("Not enough bytes to encode cell descriptors")

    d1, d2 = cell_data[0], cell_data[1]
    cell_data = cell_data[2:]
    level = math.floor(d1 / 32)
    is_exotic = d1 & 8
    ref_num = d1 % 8
    data_bytes_size = math.ceil(d2 / 2)
    fullfilled_bytes = not (d2 % 2)

    cell = Cell()
    cell.is_exotic = is_exotic

    if len(cell_data) < data_bytes_size + reference_index_size * ref_num:
        raise Exception("Not enough bytes to encode cell data")

    cell.bits.set_top_upped_array(
        bytearray(cell_data[:data_bytes_size]), fullfilled_bytes)
    cell_data = cell_data[data_bytes_size:]
    for r in range(ref_num):
        cell.refs.append(read_n_bytes_uint_from_array(
            reference_index_size, cell_data))
        cell_data = cell_data[reference_index_size:]

    return {
        "cell": cell,
        "residue": cell_data
    }


def parse_boc_header(serialized_boc):
    if len(serialized_boc) < 4 + 1:
        raise Exception("Not enough bytes for magic prefix")

    input_data = serialized_boc
    prefix = serialized_boc[:4]
    serialized_boc = serialized_boc[4:]
    if compare_bytes(prefix, Cell.REACH_BOC_MAGIC_PREFIX):
        flags_byte = serialized_boc[0]
        has_idx = flags_byte & 128
        hash_crc32 = flags_byte & 64
        has_cache_bits = flags_byte & 32
        flags = (flags_byte & 16) * 2 + (flags_byte & 8)
        size_bytes = flags_byte % 8

    if compare_bytes(prefix, Cell.LEAN_BOC_MAGIC_PREFIX):
        has_idx = 1
        hash_crc32 = 0
        has_cache_bits = 0
        flags = 0
        size_bytes = serialized_boc[0]

    if compare_bytes(prefix, Cell.LEAN_BOC_MAGIC_PREFIX_CRC):
        has_idx = 1
        hash_crc32 = 1
        has_cache_bits = 0
        flags = 0
        size_bytes = serialized_boc[0]

    serialized_boc = serialized_boc[1:]

    if len(serialized_boc) < 1 + 5 * size_bytes:
        raise Exception("Not enough bytes for encoding cells counters")

    offset_bytes = serialized_boc[0]
    serialized_boc = serialized_boc[1:]
    cells_num = read_n_bytes_uint_from_array(
        size_bytes, serialized_boc)
    serialized_boc = serialized_boc[size_bytes:]
    roots_num = read_n_bytes_uint_from_array(
        size_bytes, serialized_boc)
    serialized_boc = serialized_boc[size_bytes:]
    absent_num = read_n_bytes_uint_from_array(
        size_bytes, serialized_boc)
    serialized_boc = serialized_boc[size_bytes:]
    tot_cells_size = read_n_bytes_uint_from_array(
        offset_bytes, serialized_boc)
    serialized_boc = serialized_boc[offset_bytes:]

    if len(serialized_boc) < roots_num * size_bytes:
        raise Exception("Not enough bytes for encoding root cells hashes")

    root_list = []
    for c in range(roots_num):
        root_list.append(read_n_bytes_uint_from_array(
            size_bytes, serialized_boc))
        serialized_boc = serialized_boc[size_bytes:]

    index = False
    if has_idx:
        index = []
        if len(serialized_boc) < offset_bytes * cells_num:
            raise Exception("Not enough bytes for index encoding")
        for c in range(cells_num):
            index.append(read_n_bytes_uint_from_array(
                offset_bytes, serialized_boc))
            serialized_boc = serialized_boc[offset_bytes:]

    if len(serialized_boc) < tot_cells_size:
        raise Exception("Not enough bytes for cells data")
    cells_data = serialized_boc[:tot_cells_size]
    serialized_boc = serialized_boc[tot_cells_size:]

    if hash_crc32:
        if len(serialized_boc) < 4:
            raise Exception("Not enough bytes for crc32c hashsum")

        length = len(input_data)
        if not compare_bytes(crc32c(input_data[:length-4]), serialized_boc[:4]):
            raise Exception("Crc32c hashsum mismatch")

        serialized_boc = serialized_boc[4:]

    if len(serialized_boc):
        raise Exception("Too much bytes in BoC serialization")

    return {
        'has_idx': has_idx,
        'hash_crc32': hash_crc32,
        'has_cache_bits': has_cache_bits,
        'flags': flags,
        'size_bytes': size_bytes,
        'off_bytes': offset_bytes,
        'cells_num': cells_num,
        'roots_num': roots_num,
        'absent_num': absent_num,
        'tot_cells_size': tot_cells_size,
        'root_list': root_list,
        'index': index,
        'cells_data': cells_data,
    }


def deserialize_boc(serialized_boc):
    if type(serialized_boc) == str:
        serialized_boc = bytes.fromhex(serialized_boc)

    header = parse_boc_header(serialized_boc)
    cells_data = header['cells_data']
    cells_array = []

    for ci in range(header["cells_num"]):
        dd = deserialize_cell_data(cells_data, header["size_bytes"])
        cells_data = dd["residue"]
        cells_array.append(dd["cell"])

    for ci in reversed(range(header["cells_num"])):
        c = cells_array[ci]
        for ri in range(len(c.refs)):
            r = c.refs[ri]
            if (r < ci):
                raise Exception("Topological order is broken")
            c.refs[ri] = cells_array[r]

    root_cells = []
    for ri in header["root_list"]:
        root_cells.append(cells_array[ri])

    return root_cells



from abc import ABCMeta, abstractmethod
from collections import OrderedDict
import inquirer
from halo import Halo
import contextlib

from .._utils import echo_error, echo_success
from ..._utils import SharedObject
from .._modified_inquirer import ModifiedConsoleRender, ModifiedTheme
from .._exceptions import EscButtonPressed


class BaseSet(metaclass=ABCMeta):
    MENU_KEY = "menu_option"

    def __init__(self, ctx: SharedObject) -> None:
        self.ctx = ctx
        self._exit = False
        self._spinner = Halo(text='Processing', spinner='dots')

    def show(self):
        while not self._exit:
            items = [
                inquirer.List(self.MENU_KEY, message="Pick command",
                              choices=self._handlers().keys(), carousel=True),
            ]

            try:
                item = self._prompt(items)[self.MENU_KEY]
            except EscButtonPressed as e:
                self._handle_exit()
                continue

            try:
                self._handlers()[item]()
            except EscButtonPressed as e:
                pass
            except Exception as e:
                if self.ctx.debug_mode:
                    raise

                echo_error(e.__repr__())

    def _prompt(self, questions):
        return inquirer.prompt(questions, render=ModifiedConsoleRender(theme=ModifiedTheme()), raise_keyboard_interrupt=True)

    @abstractmethod
    def _handlers(self) -> OrderedDict:
        raise NotImplementedError

    def _handle_exit(self):
        self._exit = True

    def _start_loading(self):
        self._spinner.start()

    def _stop_loading(self):
        self._spinner.stop()

    @contextlib.contextmanager
    def _processing(self):
        try:
            self._start_loading()
            yield
        finally:
            self._stop_loading()



import base64
import ctypes
from ._utils import crc16, string_to_bytes
from ._exceptions import InvalidAddressError


def parse_friendly_address(addr_str):
    if len(addr_str) != 48:
        raise InvalidAddressError(
            "User-friendly address should contain strictly 48 characters")

    # avoid padding error (https://gist.github.com/perrygeo/ee7c65bb1541ff6ac770)
    data = string_to_bytes(base64.b64decode(addr_str+"=="))

    if len(data) != 36:
        raise InvalidAddressError(
            "Unknown address type: byte length is not equal to 36")

    addr = data[:34]
    crc = data[34:36]
    calced_crc = crc16(addr)
    if not (calced_crc[0] == crc[0] and calced_crc[1] == crc[1]):
        raise InvalidAddressError("Wrong crc16 hashsum")

    tag = addr[0]
    is_test_only = False
    is_bounceable = False
    if tag & Address.TEST_FLAG:
        is_test_only = True
        tag ^= Address.TEST_FLAG
    if (tag != Address.BOUNCEABLE_TAG) and (tag != Address.NON_BOUNCEABLE_TAG):
        raise InvalidAddressError("Unknown address tag")

    is_bounceable = tag == Address.BOUNCEABLE_TAG

    if addr[1] == 0xff:
        workchain = -1
    else:
        workchain = addr[1]
    if workchain != 0 and workchain != -1:
        raise InvalidAddressError(f"Invalid address wc {workchain}")

    hash_part = bytearray(addr[2:34])
    return {
        "is_test_only": is_test_only,
        "is_bounceable": is_bounceable,
        "workchain": workchain,
        "hash_part": hash_part,
    }


class Address:
    BOUNCEABLE_TAG = 0x11
    NON_BOUNCEABLE_TAG = 0x51
    TEST_FLAG = 0x80

    def __init__(self, any_form):
        if any_form is None:
            raise InvalidAddressError("Invalid address")

        if isinstance(any_form, Address):
            self.wc = any_form.wc
            self.hash_part = any_form.hash_part
            self.is_test_only = any_form.is_test_only
            self.is_user_friendly = any_form.is_user_friendly
            self.is_bounceable = any_form.is_bounceable
            self.is_url_safe = any_form.is_url_safe
            return

        if any_form.find("-") > 0 or any_form.find("_") > 0:
            any_form = any_form.replace("-", '+').replace("_", '/')
            self.is_url_safe = True
        else:
            self.is_url_safe = False

        try:
            colon_index = any_form.index(":")
        except ValueError:
            colon_index = -1

        if colon_index > -1:
            arr = any_form.split(":")
            if len(arr) != 2:
                raise InvalidAddressError(f"Invalid address {any_form}")

            wc = int(arr[0])
            if wc != 0 and wc != 1:
                raise InvalidAddressError(f"Invalid address wc {wc}")

            address_hex = arr[1]
            if len(address_hex) != 64:
                raise InvalidAddressError(f'Invalid address hex {any_form}')

            self.is_user_friendly = False
            self.wc = wc
            self.hash_part = bytearray.fromhex(address_hex)
            self.is_test_only = False
            self.is_bounceable = False
        else:
            self.is_user_friendly = True
            parse_result = parse_friendly_address(any_form)
            self.wc = parse_result["workchain"]
            self.hash_part = parse_result["hash_part"]
            self.is_test_only = parse_result["is_test_only"]
            self.is_bounceable = parse_result["is_bounceable"]

    def to_string(self, is_user_friendly=None, is_url_safe=None, is_bounceable=None, is_test_only=None):
        if is_user_friendly is None:
            is_user_friendly = self.is_user_friendly
        if is_url_safe is None:
            is_url_safe = self.is_url_safe
        if is_bounceable is None:
            is_bounceable = self.is_bounceable
        if is_test_only is None:
            is_test_only = self.is_test_only

        if not is_user_friendly:
            return f"{self.wc}:{self.hash_part.hex()}"
        else:
            tag = Address.BOUNCEABLE_TAG if is_bounceable else Address.NON_BOUNCEABLE_TAG

            if is_test_only:
                tag |= Address.TEST_FLAG

            addr = (ctypes.c_int8 * 34)()
            addr[0] = tag
            addr[1] = self.wc
            addr[2:] = self.hash_part
            address_with_checksum = (ctypes.c_uint8 * 36)()
            address_with_checksum[:34] = addr
            address_with_checksum[34:] = crc16(addr)

            address_base_64 = base64.b64encode(
                address_with_checksum).decode('utf-8')
            if is_url_safe:
                address_base_64 = address_base_64.replace(
                    "+", '-').replace("/", '_')

            return str(address_base_64)

    def to_buffer(self):
        return self.hash_part + bytearray([self.wc, self.wc, self.wc, self.wc])



import json
import os
from typing import Dict, List, Optional, Tuple, Union
from pydantic import BaseModel, validator
from hashlib import sha256
from nacl.bindings import crypto_box_seed_keypair, crypto_box, crypto_box_open

from wton.tonsdk.crypto import generate_new_keystore, generate_keystore_key
from wton.tonsdk.utils import Address, InvalidAddressError
from wton.utils import storage
from wton.tonsdk.contract.wallet import WalletVersionEnum, Wallets
from ._exceptions import KeyStoreAlreadyExistsError, RecordAlreadyExistsError, \
    RecordDoesNotExistError, KeyStoreDoesNotExistError, InvalidKeyStoreError, \
    KeyStoreIsNotSpecifiedError, KeyStoreInvalidPasswordError


class TonCliBackupRecord(BaseModel):
    name: str
    comment: Optional[str] = ""
    config: str
    kind: str
    address: Union[str, Address]
    mnemonics: List[str]

    class Config:
        use_enum_values = True
        validate_assignment = True
        arbitrary_types_allowed = True

    def to_backup_record(self) -> Optional["BackupRecord"]:
        if not self.__supported_wallet():
            return None

        return BackupRecord(name=self.name, address=self.address,
                            version=self.__map_kind_to_version(), workchain=self.__map_config_to_workchain(),
                            mnemonics=" ".join(self.mnemonics), comment=self.comment)

    def __supported_wallet(self):
        return self.kind in self.kind_version_map.keys()

    def __map_kind_to_version(self):
        return self.kind_version_map[self.kind]

    def __map_config_to_workchain(self):
        # "wc=0,walletId=698983191,pk=qweqweqwe"
        return self.config.split(",")[0].split("=")[1]

    @property
    def kind_version_map(self) -> Dict:
        return {
            "org.ton.wallets.v2": WalletVersionEnum.v2r1,
            "org.ton.wallets.v2.r2": WalletVersionEnum.v2r2,
            "org.ton.wallets.v3": WalletVersionEnum.v3r1,
            "org.ton.wallets.v3.r2": WalletVersionEnum.v3r2,
        }


class BackupRecord(BaseModel):
    name: str
    address: Union[str, Address]
    version: WalletVersionEnum
    workchain: int
    mnemonics: str
    comment: Optional[str] = ""

    class Config:
        use_enum_values = True
        validate_assignment = True
        arbitrary_types_allowed = True

    @classmethod
    def from_record(cls, record: "Record", mnemonics: str) -> "BackupRecord":
        return cls(
            name=record.name,
            address=record.address,
            version=record.version,
            workchain=record.workchain,
            mnemonics=mnemonics,
            comment=record.comment,
        )


class Record(BaseModel):
    name: str
    address: Union[str, Address]
    version: WalletVersionEnum
    workchain: int
    public_key: str
    secret_key: str
    comment: Optional[str] = ""

    class Config:
        use_enum_values = True
        validate_assignment = True
        arbitrary_types_allowed = True

    @validator('address')
    def validate_address(cls, v, values, **kwargs):
        if v:
            if isinstance(v, Address):
                return v.to_string(False, False, False)

            try:
                addr = Address(v)
                return addr.to_string(False, False, False)

            except InvalidAddressError as e:
                raise ValueError(e)

    @property
    def address_to_show(self):
        return Address(self.address).to_string(True, True, True)


class KeyStore(BaseModel):
    filepath: str
    salt: str = ""
    public_key: str = ""
    version: int = 1
    records: List[Record] = []

    class Config:
        use_enum_values = True
        validate_assignment = True

    @classmethod
    def new(cls, filepath: str, password: str) -> 'KeyStore':
        data = generate_new_keystore(password)
        data["filepath"] = filepath
        return cls.parse_obj(data)

    @classmethod
    def load(cls, filepath) -> 'KeyStore':
        raw_data = storage.read_bytes(filepath)
        if len(raw_data) < 32:
            raise InvalidKeyStoreError(f"Broken keystore: {filepath}")

        hash_data = raw_data[:32]
        data = raw_data[32:]
        if hash_data != sha256(data).digest():
            raise InvalidKeyStoreError(f"Broken keystore: {filepath}")

        json_data = json.loads(data.decode('utf-8'))
        json_data["filepath"] = filepath

        return cls.parse_obj(json_data)

    def save(self):
        json_data = json.dumps(self.dict(exclude={'filepath'})).encode('utf-8')
        hash_of_data = sha256(json_data).digest()
        storage.save_bytes(self.filepath, hash_of_data + json_data)

    def backup(self, filepath: str, password: str):
        backup_records = []
        for record in self.records:
            mnemonics = self.get_secret(record, password)
            backup_records.append(
                BackupRecord.from_record(record, mnemonics).dict())

        storage.save_json(filepath, backup_records)

    def get_record_by_name(self, name: str, raise_none: bool = False):
        return self.__get_record(name=name, raise_none=raise_none)

    def get_record_by_address(self, address: Union[str, Address], raise_none: bool = False):
        return self.__get_record(address=address, raise_none=raise_none)

    def edit_record(self, name: str, new_name: str, new_comment: str, save: bool = False):
        record = self.get_record_by_name(name, raise_none=True)
        record_idx = self.records.index(record)

        if new_name:
            self.records[record_idx].name = new_name

        if new_comment:
            self.records[record_idx].comment = new_comment

        if save:
            self.save()

    def delete_record(self, name: str, save: bool = False):
        record = self.get_record_by_name(name, raise_none=True)
        self.records.remove(record)

        if save:
            self.save()

    def add_record(self, name: str, address: Address, public_key: str,
                   mnemonics: List[str], version: WalletVersionEnum,
                   workchain: int, comment: Optional[str] = None,
                   password: Optional[str] = None, save=False) -> Record:
        if self.__get_record(name=name, address=address) is not None:
            raise RecordAlreadyExistsError(
                f"Record with the name '{name}' or address: '{address.to_string(True, True, True)}' already exists")

        key = (" ".join(mnemonics)).encode('utf-8')
        sk = self.__create_record_sk(key)

        new_record = Record(name=name, address=address,
                            version=version, workchain=workchain, comment=comment,
                            public_key=public_key, secret_key=sk.hex())

        self.records.append(new_record)

        if save:
            self.save()

    def get_secret(self, record: Record, password: str) -> str:
        src = bytes.fromhex(record.secret_key)
        nonce, public_key, data = src[:24], src[24:24+32], src[24+32:]

        pub_k, priv_k = generate_keystore_key(
            password, bytes.fromhex(self.salt))
        if pub_k != bytes.fromhex(self.public_key):
            raise KeyStoreInvalidPasswordError("Invalid keystore password.")

        decoded_key_bytes = crypto_box_open(data, nonce, public_key, priv_k)
        if not decoded_key_bytes:
            raise KeyStoreInvalidPasswordError("Invalid keystore password.")

        return decoded_key_bytes.decode("utf-8")

    def __create_record_sk(self, key: bytes) -> bytes:
        ephemeral_key_public, ephemeral_key_secret = crypto_box_seed_keypair(
            os.urandom(32))
        nonce = os.urandom(24)
        encrypted = crypto_box(key, nonce, bytes.fromhex(
            self.public_key), ephemeral_key_secret)

        return nonce + ephemeral_key_public + encrypted

    def __get_record(self, name: Optional[str] = None, address: Union[str, Address, None] = None,
                     raise_none: bool = False) -> Optional[Record]:
        record = None

        if name:
            record = next(
                (record for record in self.records if record.name == name), record)
            if record is None and raise_none:
                raise RecordDoesNotExistError(
                    f"Record with the name {name} does not exist")

        if address:
            address = address if isinstance(
                address, str) else address.to_string(False, False, False)
            record = next(
                (record for record in self.records if record.address == address), record)
            if record is None and raise_none:
                raise RecordDoesNotExistError(
                    f"Record with the address {address} does not exist")

        return record


class KeyStores:
    def __init__(self, keystores_workdir: str):
        self.keystores_workdir = keystores_workdir

        paths = storage.get_filenames_by_ptrn(
            keystores_workdir, "*.keystore")
        self.keystore_paths = {os.path.basename(path): path for path in paths}
        self._keystore_path = None

    def create_new_keystore(self, name: str, password: Optional[str] = None, save: bool = False) -> Tuple[KeyStore, str]:
        name = self.__keystore_name(name)

        if name in self.keystore_paths:
            raise KeyStoreAlreadyExistsError(
                f"Keystore with the name '{name}' already exists")

        if password is None:
            # TODO: generate password
            raise NotImplemented("Empty password is not supported")

        filepath = os.path.join(self.keystores_workdir, name)
        keystore = KeyStore.new(filepath=filepath, password=password)
        if save:
            keystore.save()

        self.keystore_paths[name] = keystore.filepath

        return keystore, password

    def get_keystore(self, keystore_name: Optional[str], raise_none: bool = False) -> Optional[KeyStore]:
        if keystore_name is None:
            raise KeyStoreIsNotSpecifiedError(
                "wton.keystore_name is not specified.")

        keystore_path = os.path.join(self.keystores_workdir, keystore_name)
        try:
            return KeyStore.load(keystore_path)

        except FileNotFoundError:
            if raise_none:
                raise KeyStoreDoesNotExistError(
                    f"Keystore with the name '{keystore_name}' does not exist.")

            return None

        except json.JSONDecodeError as e:
            raise InvalidKeyStoreError("Invalid keystore file. " + str(e))

    def restore_wton_keystore(self, name: str, filepath: str, password: Optional[str] = None):
        backup_records: List[BackupRecord] = []
        backup_raw_records = storage.read_json(filepath)
        for raw_record in backup_raw_records:
            backup_records.append(
                BackupRecord.parse_obj(raw_record))

        self.__restore_keystore(name, filepath, backup_records, password)

    def restore_ton_cli_keystore(self, name: str, filepath: str, password: Optional[str] = None):
        backup_records: List[BackupRecord] = []
        backup_ton_cli_raw_records = storage.read_json(filepath)
        for raw_record in backup_ton_cli_raw_records:
            backup_record = TonCliBackupRecord.parse_obj(
                raw_record).to_backup_record()
            if backup_record:
                backup_records.append(backup_record)

        self.__restore_keystore(name, filepath, backup_records, password)

    def __restore_keystore(self, name: str, filepath: str, backup_records: List[BackupRecord], password: Optional[str] = None):
        name = self.__keystore_name(name)

        keystore, _password = self.create_new_keystore(
            name, password, save=False)

        for backup_record in backup_records:
            mnemonics, pub_k, _priv_k, wallet = Wallets.from_mnemonics(
                backup_record.mnemonics.split(
                    " "), backup_record.version,
                backup_record.workchain)

            keystore.add_record(backup_record.name, wallet.address,
                                pub_k.hex(), mnemonics, backup_record.version,
                                backup_record.workchain, backup_record.comment,
                                save=False)
        keystore.save()

        self.keystore_paths[name] = keystore.filepath

    def __keystore_name(self, name):
        if not name.endswith(".keystore"):
            name += ".keystore"
        return name



import os
import click

from wton.utils import storage

from ._base_cmd import cli, SharedObject
from .._utils import CustomClickException, with_whitelist, with_keystore, click_ton_exception_handler, click_echo_success
from ..._utils import md_table
from wton.tonsdk.contract.wallet import SendModeEnum, WalletVersionEnum, Wallets
from wton.tonclient.utils import RecordAlreadyExistsError, RecordDoesNotExistError,\
    WhitelistContactDoesNotExistError, KeyStoreInvalidPasswordError, InvalidMnemonicsError
from wton.tonsdk.utils import TonCurrencyEnum, Address
from wton.tonclient import ton_exceptions_handler


@cli.group()
def wallet():
    """
    Operate with wallets
    """


@wallet.command()
@with_keystore(password_required=False)
@with_whitelist
@click.argument('name', required=True)
@click.option('--version', '-v', type=click.Choice(WalletVersionEnum))
@click.option('--workchain', '-wc', default=0, show_default=True, type=int)
@click.option('--comment', help='Extra information about the wallet.')
@click.option('--save-to-whitelist', 'contact_name', help='Contact name to save.', metavar='NAME')
@click.pass_obj
def create(shared_object: SharedObject, name: str, version: str, workchain: int, comment: str, contact_name: str):
    """
    Create wallet data and add it to the keystore.
    """
    if contact_name:
        contact = shared_object.whitelist.get_contact(contact_name)
        if contact is not None:
            raise CustomClickException(
                f"Contact with the name '{contact_name}' already exists")

    if not version:
        version = WalletVersionEnum(
            shared_object.config.wton.default_wallet_version)

    try:
        mnemonics, pub_k, _priv_k, wallet = Wallets.create(version, workchain)
        shared_object.keystore.add_record(name, wallet.address,
                                          pub_k.hex(), mnemonics, version,
                                          workchain, comment, save=True)
    except RecordAlreadyExistsError as e:
        raise CustomClickException(e)

    if contact_name:
        shared_object.whitelist.add_contact(
            contact_name, wallet.address.to_string(True), save=True)

    click_echo_success(f"wallet {name} has been created.")

@wallet.command()
@with_keystore(password_required=False)
@with_whitelist
@click.argument('name', required=True)
@click.option('--name', '-n', 'new_name', help='New wallet name.')
@click.option('--comment', '-c', 'new_comment', help='New extra information about the wallet.')
@click.pass_obj
def edit(shared_object: SharedObject, name: str, new_name: str, new_comment: str):
    """
    Edit wallet data in a keystore.
    """
    try:
        shared_object.keystore.edit_record(
            name, new_name, new_comment, save=True)
    except RecordDoesNotExistError as e:
        raise CustomClickException(e)

    click_echo_success(f"wallet {name} has been edited.")

@wallet.command()
@with_keystore(password_required=False)
@with_whitelist
@click.argument('name', required=True)
@click.option('--yes', '-y', 'is_sure', is_flag=True, help='Do not show the prompt.')
@click.pass_obj
def delete(shared_object: SharedObject, name: str, is_sure: bool):
    """
    Delete wallet data from a keystore.
    """
    if not is_sure:
        click.confirm(
            f'Are you sure you want to delete {name} wallet?', abort=True)

    try:
        shared_object.keystore.delete_record(name, save=True)
    except RecordDoesNotExistError as e:
        raise CustomClickException(e)

    click_echo_success(f"wallet {name} has been deleted.")

@wallet.command()
@with_keystore(password_required=False)
@with_whitelist
@click.argument('name', required=True)
@click.option('--verbose', '-v', is_flag=True, help='Load info about wallet from TON network.')
@click.pass_obj
def get(shared_object: SharedObject, name: str, verbose: bool):
    """
    Get wallet data from a keystore.
    """
    try:
        wallet = shared_object.keystore.get_record_by_name(
            name, raise_none=True)
    except RecordDoesNotExistError as e:
        raise CustomClickException(e)

    addr = Address(wallet.address)
    if verbose:
        addr_info = shared_object.ton_client.get_address_information(
            wallet.address)

    click.echo(f"Raw address: {addr.to_string(False, False, False)}")
    click.echo(f"Nonbounceable address: {addr.to_string(True, True, False)}")
    click.echo(f"Bounceable address: {addr.to_string(True, True, True)}")
    click.echo(f"Version: {wallet.version}")
    click.echo(f"Workchain: {wallet.workchain}")
    click.echo(f"Comment: {wallet.comment}")

    if verbose:
        click.echo("--- Verbose wallet information ---")
        for k, v in addr_info.dict().items():
            click.echo(str(k) + ': ' + str(v))


@wallet.command()
@ton_exceptions_handler(click_ton_exception_handler)
@with_keystore(password_required=False)
@click.option('--currency', '-c', default=TonCurrencyEnum.ton, show_default=True, type=click.Choice(TonCurrencyEnum))
@click.option('--verbose', '-v', 'verbose', is_flag=True, default=False, help="Extra information from TON.")
@click.pass_obj
def list(shared_object: SharedObject, verbose: bool, currency: TonCurrencyEnum):
    """
    Print all wallets info as a markdown table. 
    You can output this command into .md file.
    """
    field_names = ['Name', 'Version', 'WC', 'Address', 'Comment']
    if verbose:
        field_names += ['State', 'Balance']

    wallets = shared_object.keystore.records

    table = md_table()
    table.field_names = field_names
    if verbose:
        wallet_infos = shared_object.ton_client.get_addresses_information(
            [wallet.address for wallet in wallets], currency)

        for wallet, wallet_info in zip(wallets, wallet_infos):
            table.add_row([wallet.name, wallet.version, wallet.workchain,
                           wallet.address_to_show, wallet.comment,
                           wallet_info.state, wallet_info.balance])
    else:
        for wallet in wallets:
            table.add_row([wallet.name, wallet.version, wallet.workchain,
                           wallet.address_to_show, wallet.comment])

    click.echo(table)


@wallet.command()
@with_keystore(password_required=False)
@with_whitelist
@click.argument('name', required=True)
@click.argument('version', type=click.Choice(WalletVersionEnum))
@click.argument('workchain', type=int)
@click.argument('mnemonics')
@click.option('--comment', help='Extra information about the wallet.')
@click.option('--save-to-whitelist', 'contact_name', help='Contact name to save.', metavar='NAME')
@click.pass_obj
def import_from_mnemonics(shared_object: SharedObject, name: str, version: str, workchain: int,
                          mnemonics: str, comment: str, contact_name: str):
    """
    Create wallet data from mnemonics and add it to the keystore.
    """
    if contact_name:
        contact = shared_object.whitelist.get_contact(contact_name)
        if contact is not None:
            raise CustomClickException(
                f"Contact with the name '{contact_name}' already exists")

    mnemonics = mnemonics.split(" ")
    try:
        mnemonics, pub_k, _priv_k, wallet = Wallets.from_mnemonics(
            mnemonics, version, workchain)
        shared_object.keystore.add_record(name, wallet.address,
                                          pub_k.hex(), mnemonics, version,
                                          workchain, comment, save=True)
    except (RecordAlreadyExistsError, InvalidMnemonicsError) as e:
        raise CustomClickException(e)

    if contact_name:
        shared_object.whitelist.add_contact(
            contact_name, wallet.address.to_string(True), save=True)

    click_echo_success(f"wallet {name} has been imported.")


@wallet.command()
@with_keystore(password_required=True)
@click.argument('name', required=True)
@click.pass_obj
def reveal(shared_object: SharedObject, name: str):
    """
    Echo mnemonics of a wallet by its name.
    """
    try:
        record = shared_object.keystore.get_record_by_name(
            name, raise_none=True)
        mnemonics = shared_object.keystore.get_secret(
            record, shared_object.keystore_password)
        click.echo(mnemonics)

    except (RecordDoesNotExistError, KeyStoreInvalidPasswordError) as e:
        raise CustomClickException(e)


@wallet.command()
@with_keystore(password_required=True)
@click.argument('name', required=True)
@click.argument('destination_dir', default=".", required=False, metavar='DESTINATION_DIR')
@click.pass_obj
def to_addr_pk(shared_object: SharedObject, name: str, destination_dir: str):
    """
    Export wallet to .pk and .addr file into a specified directory.
    """
    try:
        record = shared_object.keystore.get_record_by_name(
            name, raise_none=True)
        mnemonics = shared_object.keystore.get_secret(
            record, shared_object.keystore_password).split(" ")
        addr, pk = Wallets.to_addr_pk(mnemonics, record.version,
                                      record.workchain)

        addr_path = os.path.join(
            destination_dir, record.name + ".addr")
        pk_path = os.path.join(destination_dir, record.name + ".pk")
        storage.save_bytes(addr_path, addr)
        storage.save_bytes(pk_path, pk)

    except (RecordDoesNotExistError, KeyStoreInvalidPasswordError, OSError) as e:
        raise CustomClickException(e)


@wallet.command()
@ton_exceptions_handler(click_ton_exception_handler)
@with_whitelist
@with_keystore(password_required=True)
@click.argument('name', required=True)
@click.pass_obj
def init(shared_object: SharedObject, name: str):
    """
    Initialize address as a wallet.
    """
    try:
        record = shared_object.keystore.get_record_by_name(
            name, raise_none=True)
        mnemonics = shared_object.keystore.get_secret(
            record, shared_object.keystore_password).split(" ")
    except (RecordDoesNotExistError, KeyStoreInvalidPasswordError) as e:
        raise CustomClickException(e)

    _mnemonics, _pub_k, _priv_k, wallet = Wallets.from_mnemonics(mnemonics, record.version,
                                                                 record.workchain)

    result = shared_object.ton_client.deploy_wallet(wallet)

    click_echo_success(result)


@wallet.command()
@ton_exceptions_handler(click_ton_exception_handler)
@with_whitelist
@with_keystore(password_required=True)
@click.argument('from_wallet', required=True, metavar='WALLET_NAME')
@click.argument('to_contact', required=True, metavar='CONTACT_NAME')
@click.argument('amount', required=False, metavar='TON_COINS_NUM')
@click.option('--message', '-m', help='Attach message to the transfer.')
@click.option('--wait', '-w', is_flag=True, help='Wait until transaction is committed.', default=False)
@click.option('--ignore-errors', default="y", type=click.Choice(["y", "n"]), show_default=True)
@click.option('--pay-gas-separately', default="y", type=click.Choice(["y", "n"]), show_default=True)
@click.option('--destroy-if-zero', default="n", type=click.Choice(["y", "n"]), show_default=True)
@click.option('--transfer-all', default="n", type=click.Choice(["y", "n"]), show_default=True)
@click.pass_obj
def transfer(shared_object: SharedObject, from_wallet, to_contact, amount, message, wait,
             ignore_errors, pay_gas_separately, destroy_if_zero, transfer_all):
    """
    Transfer coins from your wallet to any address.
    """
    if amount is None and not transfer_all:
        raise CustomClickException(
            "You must specify amount when you do not use --transfer-all flag.")

    try:
        record = shared_object.keystore.get_record_by_name(
            from_wallet, raise_none=True)
        mnemonics = shared_object.keystore.get_secret(
            record, shared_object.keystore_password).split(" ")
        contact = shared_object.whitelist.get_contact(
            to_contact, raise_none=True)
    except (WhitelistContactDoesNotExistError, RecordDoesNotExistError, KeyStoreInvalidPasswordError) as e:
        raise CustomClickException(e)

    send_mode = 0
    if ignore_errors == "y":
        send_mode |= SendModeEnum.ignore_errors
    if pay_gas_separately == "y":
        send_mode |= SendModeEnum.pay_gas_separately
    if destroy_if_zero == "y":
        send_mode |= SendModeEnum.destroy_account_if_zero
    if transfer_all == "y":
        send_mode |= SendModeEnum.carry_all_remaining_balance

    _mnemonics, _pub_k, _priv_k, wallet = Wallets.from_mnemonics(mnemonics, record.version,
                                                                 record.workchain)
    result = shared_object.ton_client.transfer(
        wallet, contact.address, amount, message, send_mode, wait)

    click_echo_success(result)



from .._utils import init_shared_object, setup_app


def main():
    # config = init_shared_object(specific_config_path)
    # setup_app(config)
    raise SystemExit("Gui version hasn't been implemented yet")



"""
https://ton-blockchain.github.io/docs/tvm.pdf
4.5.7. List of predefined exceptions
"""
import functools


class TonError(Exception):
    """
    Base class for ton exceptions.
    Subclasses should provide `.status_code` and `.default_detail` properties.
    """
    default_detail = 'TON error.'
    default_code = 500

    def __init__(self, detail=None, code=None):
        self.code = self.default_code if code is None else code
        self.detail = self.default_detail if detail is None else detail

        error_msg = "TON error."
        if self.code:
            error_msg += f" Code: {self.code}."
        if self.detail:
            error_msg += f" {self.detail}"
        self.error_msg = error_msg

    def __str__(self):
        return self.error_msg


class TonContractUninitializedError(TonError):
    default_detail = "Contract is not initialized."
    default_code = -13


class TonStackUnderflowError(TonError):
    default_detail = "Not enough arguments in the stack for a primitive."
    default_code = 2


class TonStackOverflowError(TonError):
    default_detail = "More values have been stored on a stack than allowed by this version of TVM."
    default_code = 3


class TonIntegerOverflowError(TonError):
    default_detail = "Integer does not fit into -(2^256) ≤ x < 2^256, or a division by zero has occurred."
    default_code = 4


class TonRangeCheckError(TonError):
    default_detail = "Integer out of expected range."
    default_code = 5


class TonInvalidOpcodeError(TonError):
    default_detail = "Instruction or its immediate arguments cannot be decoded."
    default_code = 6


class TonTypeCheckError(TonError):
    default_detail = "An argument to a primitive is of incorrect value type."
    default_code = 7


class TonCellOverflowError(TonError):
    default_detail = "Error in one of the serialization primitives."
    default_code = 8


class TonCellUnderflowError(TonError):
    default_detail = "Deserialization error."
    default_code = 9


class TonDictionaryError(TonError):
    default_detail = "Error while deserializing a dictionary object."
    default_code = 10


class TonUnknownError(TonError):
    default_detail = "Unknown error, may be thrown by user programs."
    default_code = 11


class TonFatalError(TonError):
    default_detail = "Thrown by TVM in situations deemed impossible."
    default_code = 12


class TonOutOfGasError(TonError):
    default_detail = "Thrown by TVM when the remaining gas (gr) becomes negative. This exception usually cannot be caught and leads to an immediate termination of TVM."
    default_code = 13


class TonNetworkError(TonError):
    default_detail = "Network problems. It may be caused by TON itself (e.g. testnet update). Make sure you have an internet connection and TON config is correct."
    default_code = 500


class TonTooManyQueriesError(TonError):
    default_detail = "Too many queries error."
    default_code = 429


class TonUndocumentedError(TonError):
    default_detail = "Undocummented TON error. wton does not know what it is."
    default_code = None


class TonTimeoutError(TonError):
    default_detail = "TON request timeout error."
    default_code = None


class TonDappError(TonError):
    default_detail = "TON Dapp error."
    default_code = None


def ton_exceptions_handler(callback_by_exception):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except tuple(exc for exc in TON_EXCEPTION_BY_CODE.values()) as e:
                callback_by_exception(e)
            except (TonTimeoutError, TonUndocumentedError, TonDappError) as e:
                callback_by_exception(e)

        return wrapper

    return decorator


# TODO: separate TON network and Dapp exceptions
TON_EXCEPTION_BY_CODE = {
    -13: TonContractUninitializedError,
    2: TonStackUnderflowError,
    3: TonStackOverflowError,
    4: TonIntegerOverflowError,
    5: TonRangeCheckError,
    6: TonInvalidOpcodeError,
    7: TonTypeCheckError,
    8: TonCellOverflowError,
    9: TonCellUnderflowError,
    10: TonDictionaryError,
    11: TonUnknownError,
    12: TonFatalError,
    13: TonOutOfGasError,
    429: TonTooManyQueriesError,
    # TODO: 500 looks like "cannot apply external message to current state : Failed to unpack account state"
    500: TonNetworkError,
}



import base64


bounceable_tag, non_bounceable_tag = b'\x11', b'\x51'
b64_abc = set(
    'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890+/')
b64_abc_urlsafe = set(
    'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890_-')


def is_int(x):
    try:
        int(x)
        return True
    except:
        return False


def is_hex(x):
    try:
        int(x, 16)
        return True
    except:
        return False


def calcCRC(message):
    poly = 0x1021
    reg = 0
    message += b'\x00\x00'
    for byte in message:
        mask = 0x80
        while(mask > 0):
            reg <<= 1
            if byte & mask:
                reg += 1
            mask >>= 1
            if reg > 0xffff:
                reg &= 0xffff
                reg ^= poly
    return reg.to_bytes(2, "big")


def account_forms(raw_form, test_only=False):
    workchain, address = raw_form.split(":")
    workchain, address = int(workchain), int(address, 16)
    address = address.to_bytes(32, "big")
    workchain_tag = b'\xff' if workchain == - \
        1 else workchain.to_bytes(1, "big")
    btag = bounceable_tag
    nbtag = non_bounceable_tag
    # if test_only:
    #  btag = (btag[0] | 0x80).to_bytes(1,'big')
    #  nbtag = (nbtag[0] | 0x80).to_bytes(1,'big')
    preaddr_b = btag + workchain_tag + address
    preaddr_u = nbtag + workchain_tag + address
    b64_b = base64.b64encode(preaddr_b+calcCRC(preaddr_b)).decode('utf8')
    b64_u = base64.b64encode(preaddr_u+calcCRC(preaddr_u)).decode('utf8')
    b64_b_us = base64.urlsafe_b64encode(
        preaddr_b+calcCRC(preaddr_b)).decode('utf8')
    b64_u_us = base64.urlsafe_b64encode(
        preaddr_u+calcCRC(preaddr_u)).decode('utf8')
    return {'raw_form': raw_form,
            'bounceable': {'b64': b64_b, 'b64url': b64_b_us},
            'non_bounceable': {'b64': b64_u, 'b64url': b64_u_us},
            'given_type': 'raw_form',
            'test_only': test_only}


def read_friendly_address(address):
    urlsafe = False
    if set(address).issubset(b64_abc):
        address_bytes = base64.b64decode(address.encode('utf8'))
    elif set(address).issubset(b64_abc_urlsafe):
        urlsafe = True
        address_bytes = base64.urlsafe_b64decode(address.encode('utf8'))
    else:
        raise Exception("Not an address")
    if not calcCRC(address_bytes[:-2]) == address_bytes[-2:]:
        raise Exception("Wrong checksum")
    tag = address_bytes[0]
    if tag & 0x80:
        test_only = True
        tag = tag ^ 0x80
    else:
        test_only = False
    tag = tag.to_bytes(1, 'big')
    if tag == bounceable_tag:
        bounceable = True
    elif tag == non_bounceable_tag:
        bounceable = False
    else:
        raise Exception("Unknown tag")
    if address_bytes[1:2] == b'\xff':
        workchain = -1
    else:
        workchain = address_bytes[1]
    hx = hex(int.from_bytes(address_bytes[2:-2], "big"))[2:]
    hx = (64-len(hx))*"0"+hx
    raw_form = str(workchain)+":"+hx
    account = account_forms(raw_form, test_only)
    account['given_type'] = "friendly_" + \
        ("bounceable" if bounceable else "non_bounceable")
    return account


def detect_address(unknown_form):
    if is_hex(unknown_form):
        return account_forms("-1:"+unknown_form)
    elif (":" in unknown_form) and is_int(unknown_form.split(":")[0]) and is_hex(unknown_form.split(":")[1]):
        return account_forms(unknown_form)
    else:
        return read_friendly_address(unknown_form)


def prepare_address(unknown_form):
    address = detect_address(unknown_form)
    if 'non_bounceable' in address['given_type']:
        return address["non_bounceable"]["b64"]
    return address["bounceable"]["b64"]


def address_state(account_info):
    if isinstance(account_info.get("code", ""), int) or len(account_info.get("code", "")) == 0:
        if len(account_info.get("frozen_hash", "")) == 0:
            return "uninitialized"
        else:
            return "frozen"
    return "active"



from json import JSONDecodeError
import click
from pydantic.error_wrappers import ValidationError
import importlib.metadata

from .._utils import CustomClickException
from ..._utils import init_shared_object, SharedObject, setup_app
from wton import settings
from wton.config import Config, ConfigNotFoundError, WtonConfig, ProviderConfig
from wton.utils import storage

__version__ = importlib.metadata.version('wton')


@click.group(context_settings=dict(help_option_names=['-h', '--help']))
@click.version_option(__version__)
@click.option("-c", "--config", 'specific_config_path', metavar='', help="Use specific config.yaml file.")
@click.pass_context
def cli(ctx, specific_config_path: str):
    try:
        ctx.obj = init_shared_object(specific_config_path)

        if ctx.invoked_subcommand not in ["config", "init"]:
            setup_app(ctx.obj.config)

    except (FileNotFoundError, JSONDecodeError, ConfigNotFoundError, ValidationError) as e:
        raise CustomClickException(e)


@cli.command()
def init():
    """
    Initialize .wton workdir in a current directory.
    """
    wton = WtonConfig(workdir=settings.DEFAULT_LOCAL_WORKDIR)
    provider = ProviderConfig()
    config = Config(wton=wton, provider=provider)
    setup_app(config)
    storage.save_yaml(settings.INIT_LOCAL_CONFIG_PATH, config.dict())



from hashlib import new
import os
from collections import OrderedDict
import inquirer

from wton.tonclient.utils import Whitelist
from wton.tonsdk.contract.wallet import SendModeEnum, WalletVersionEnum, Wallets
from wton.utils import storage
from ._base import BaseSet
from ..._utils import SharedObject, md_table
from .._utils import echo_success, echo_error
from wton.tonsdk.utils import Address


class KeystoreSet(BaseSet):
    def __init__(self, ctx: SharedObject, keystore_name: str) -> None:
        super().__init__(ctx)
        ctx.keystore = ctx.keystores.get_keystore(
            keystore_name, raise_none=True)
        ctx.whitelist = Whitelist(
            ctx.config.wton.whitelist_path)

    def _handlers(self) -> OrderedDict:
        ord_dict = OrderedDict()
        ord_dict["List wallets"] = self._handle_list_wallets
        ord_dict["Transfer"] = self._handle_transfer
        ord_dict["Create wallet"] = self._handle_create_wallet
        ord_dict["Init wallet"] = self._handle_init_wallet
        ord_dict["Get wallet"] = self._handle_get_wallet
        ord_dict["Edit wallet"] = self._handle_edit_wallet
        ord_dict["Delete wallet"] = self._handle_delete_wallet
        ord_dict["Reveal wallet mnemonics"] = self._handle_reveal_wallet_mnemonics
        ord_dict["Import from mnemonics"] = self._handle_import_from_mnemonics
        ord_dict["Wallet to .addr and .pk"] = self._handle_wallet_to_addr_pk
        ord_dict["Backup keystore"] = self._handle_backup_keystore
        ord_dict["Back"] = self._handle_exit
        return ord_dict

    def _handle_list_wallets(self):
        questions = [
            inquirer.Confirm(
                "verbose", message='Show verbose information?', default=True),
        ]
        verbose = self._prompt(questions)["verbose"]

        with self._processing():
            field_names = ['Name', 'Version', 'WC', 'Address', 'Comment']
            if verbose:
                field_names += ['State', 'Balance']

            wallets = self.ctx.keystore.records

            table = md_table()
            table.field_names = field_names
            if verbose:
                wallet_infos = self.ctx.ton_client.get_addresses_information(
                    [wallet.address for wallet in wallets])

                for wallet, wallet_info in zip(wallets, wallet_infos):
                    table.add_row([wallet.name, wallet.version, wallet.workchain,
                                wallet.address_to_show, wallet.comment,
                                wallet_info.state, wallet_info.balance])
            else:
                for wallet in wallets:
                    table.add_row([wallet.name, wallet.version, wallet.workchain,
                                wallet.address_to_show, wallet.comment])

        echo_success(table, only_msg=True)

    def _handle_transfer(self):
        from_wallet = self.__select_wallet_or_false("Transfer from")
        if from_wallet == False:
            return
        questions = [
            inquirer.Password("keystore_password",
                              message='Keystore password'),
        ]
        keystore_password = self._prompt(questions)["keystore_password"]
        record = self.ctx.keystore.get_record_by_name(
            from_wallet, raise_none=True)
        mnemonics = self.ctx.keystore.get_secret(
            record, keystore_password).split(" ")

        to_contact = self.__select_contact_or_false("Send to")
        if to_contact == False:
            return
        contact = self.ctx.whitelist.get_contact(
            to_contact, raise_none=True)

        questions = [
            inquirer.Text("amount", message='Amount to transfer (TON coins)'),
            inquirer.Text(
                "message", message='Message (press \'Enter\' to skip)'),
            inquirer.Confirm(
                "destroy_if_zero", message='Destroy if zero?', default=False),
            inquirer.Confirm(
                "transfer_all", message='Transfer all?', default=False),
            inquirer.Confirm(
                "wait_for_result", message="Wait until transaction will be completed?", default=True),
        ]
        ans = self._prompt(questions)
        amount = float(ans["amount"])
        message = ans["message"]
        destroy_if_zero = ans["destroy_if_zero"]
        transfer_all = ans["transfer_all"]
        wait_for_result = ans["wait_for_result"]
        send_mode = SendModeEnum.ignore_errors | SendModeEnum.pay_gas_separately
        if destroy_if_zero:
            send_mode |= SendModeEnum.destroy_account_if_zero
        if transfer_all:
            send_mode |= SendModeEnum.carry_all_remaining_balance

        with self._processing():
            _mnemonics, _pub_k, _priv_k, wallet = Wallets.from_mnemonics(mnemonics, record.version,
                                                                         record.workchain)
            # TonDappError('timeout, Code: 0')
            result = self.ctx.ton_client.transfer(
                wallet, contact.address, amount, message, send_mode, wait_for_result)

        echo_success(result)

    def _handle_create_wallet(self):
        questions = [
            inquirer.List(
                "single", message="Choose option", choices=["Single", "Batch"],
                carousel=True),
            inquirer.List(
                "version", message='Wallet version', choices=[e.value for e in WalletVersionEnum],
                carousel=True, default=self.ctx.config.wton.default_wallet_version),
            inquirer.Text(
                "workchain", message='Workchain', default="0"),
        ]
        ans = self._prompt(questions)
        is_single = ans["single"] == "Single"
        version = WalletVersionEnum(ans["version"])
        workchain = int(ans["workchain"])

        if is_single:
            questions = [
                inquirer.Text("name", message='Wallet name'),
                inquirer.Text(
                    "comment", message='Wallet description (leave blank to skip)'),
            ]
            ans = self._prompt(questions)
            name = ans["name"]
            comment = ans["comment"]
            questions = [
                inquirer.Text(
                    "contact_name", message='Enter name to save to whitelist (blank to skip)', default=name),
            ]
            contact_name = self._prompt(questions)["contact_name"]

            wallets_to_create = [(name, comment, contact_name)]

        else:
            questions = [
                inquirer.Text("number_of_wallets",
                              message="Number of wallets to create"),
                inquirer.Text(
                    "prefix", message='Wallet name prefix (e.g. employee)'),
                inquirer.Text(
                    "comment", message='Overall wallets description (leave blank to skip)'),
                inquirer.Confirm(
                    "add_all_to_whitelist", message='Add all wallets to whitelist?', default=True),
            ]
            ans = self._prompt(questions)
            number_of_wallets = int(ans["number_of_wallets"])
            prefix = ans["prefix"] + "_"
            comment = ans["comment"]
            add_all_to_whitelist = ans["add_all_to_whitelist"]

            num_of_existing = sum(record.name.startswith(prefix)
                                  for record in self.ctx.keystore.records)
            existing_leading_zeros = len(str(num_of_existing))
            new_leading_zeros = len(str(number_of_wallets + num_of_existing))

            if new_leading_zeros != existing_leading_zeros:
                zeros_dif = new_leading_zeros - existing_leading_zeros
                replace_with = "_" + ("0" * zeros_dif)
                for record in self.ctx.keystore.records:
                    if record.name.startswith(prefix):
                        record.name = record.name.replace("_", replace_with)
                for contact in self.ctx.whitelist.contacts:
                    if contact.name.startswith(prefix):
                        contact.name = contact.name.replace("_", replace_with)

            wallets_to_create = []
            i = 1
            while len(wallets_to_create) < number_of_wallets:
                wallet_name_unique = False
                new_name = None
                contact_name = ""
                while not wallet_name_unique:
                    new_name = f"{prefix}{str(i).zfill(new_leading_zeros)}"
                    i += 1
                    if add_all_to_whitelist:
                        contact_name = new_name
                        if self.ctx.whitelist.get_contact(contact_name) is not None:
                            continue
                    if self.ctx.keystore.get_record_by_name(new_name) is not None:
                        continue
                    wallet_name_unique = True

                wallets_to_create.append(
                    (new_name, comment, contact_name))

        with self._processing():
            for wallet_name, comment, contact_name in wallets_to_create:
                if contact_name:
                    contact = self.ctx.whitelist.get_contact(contact_name)
                    if contact is not None:
                        raise Exception(
                            f"Contact with the name '{contact_name}' already exists")

                mnemonics, pub_k, _priv_k, wallet = Wallets.create(
                    version, workchain)
                self.ctx.keystore.add_record(wallet_name, wallet.address,
                                            pub_k.hex(), mnemonics, version,
                                            workchain, comment, save=False)

                if contact_name:
                    self.ctx.whitelist.add_contact(
                        contact_name, wallet.address.to_string(True), save=False)

            self.ctx.keystore.save()
            self.ctx.whitelist.save()

        echo_success()

    def _handle_init_wallet(self):
        wallet_name_to_init = self.__select_wallet_or_false("Wallet to init")
        if wallet_name_to_init == False:
            return
        questions = [
            inquirer.Confirm(
                "wait_for_result", message="Wait until transaction will be completed?", default=True),
            inquirer.Password("keystore_password",
                              message='Keystore password'),
        ]
        ans = self._prompt(questions)
        wait_for_result = ans["wait_for_result"]
        keystore_password = ans["keystore_password"]
        record = self.ctx.keystore.get_record_by_name(
            wallet_name_to_init, raise_none=True)
        mnemonics = self.ctx.keystore.get_secret(
            record, keystore_password).split(" ")
        _mnemonics, _pub_k, _priv_k, wallet = Wallets.from_mnemonics(mnemonics, record.version,
                                                                     record.workchain)
        result = self.ctx.ton_client.deploy_wallet(wallet, wait_for_result)

        if wait_for_result:
            self._parse_broadcast_result(result)
        else:
            echo_success()

    def _handle_get_wallet(self):
        wallet_name = self.__select_wallet_or_false("Get wallet")
        if wallet_name == False:
            return

        wallet = self.ctx.keystore.get_record_by_name(
            wallet_name, raise_none=True)

        questions = [
            inquirer.Confirm(
                "verbose", message='Show balances?', default=True),
        ]
        verbose = self._prompt(questions)["verbose"]

        addr = Address(wallet.address)

        if verbose:
            addr_info = self.ctx.ton_client.get_address_information(
                wallet.address)

        echo_success(
            f"Raw address: {addr.to_string(False, False, False)}", True)
        echo_success(
            f"Nonbounceable address: {addr.to_string(True, True, False)}", True)
        echo_success(
            f"Bounceable address: {addr.to_string(True, True, True)}", True)
        echo_success(f"Version: {wallet.version}", True)
        echo_success(f"Workchain: {wallet.workchain}", True)
        echo_success(f"Comment: {wallet.comment}", True)

        if verbose:
            echo_success("--- Verbose wallet information ---", True)
            for k, v in addr_info.dict().items():
                echo_success(str(k) + ': ' + str(v), True)

    def _handle_edit_wallet(self):
        wallet_name = self.__select_wallet_or_false("Edit wallet")
        if wallet_name == False:
            return

        questions = [
            inquirer.Text(
                "new_name", message='New wallet name (leave blank to skip)'),
            inquirer.Text(
                "new_comment", message='New wallet description (leave blank to skip)'),
        ]
        ans = self._prompt(questions)

        new_name = ans["new_name"]
        new_comment = ans["new_comment"]

        self.ctx.keystore.edit_record(
            wallet_name, new_name, new_comment, save=True)

        echo_success()

    def _handle_delete_wallet(self):
        questions = [
            inquirer.List(
                "single", message="Choose option", choices=["Single", "Batch"],
                carousel=True),
        ]
        is_single = self._prompt(questions)["single"] == "Single"
        if is_single:
            wallet_name = self.__select_wallet_or_false("Delete wallet")
            if wallet_name == False:
                return
            confirm_phrase = f'Are you sure you want to delete {wallet_name} wallet?'
            names_to_delete = [wallet_name]

        else:
            questions = [
                inquirer.Text(
                    "prefix", message="Etner wallet prefix to delete (e.g. employee)"),
            ]
            prefix = self._prompt(questions)["prefix"] + "_"
            confirm_phrase = f'Are you sure you want to delete all wallets with {prefix} prefix?'
            names_to_delete = [
                record.name for record in self.ctx.keystore.records if record.name.startswith(prefix)]

        questions = [
            inquirer.Confirm(
                "is_sure", message=confirm_phrase, default=False,),
            inquirer.Confirm(
                "delete_whitelist", message="Do you want to delete whitelist contacts with the same names?", default=False)
        ]
        ans = self._prompt(questions)
        is_sure = ans["is_sure"]
        delete_whitelist = ans["delete_whitelist"]

        if not is_sure:
            echo_success("Action canceled.", True)
            return

        for name in names_to_delete:
            self.ctx.keystore.delete_record(name, save=False)
            if delete_whitelist:
                self.ctx.whitelist.delete_contact(name, save=False)

        self.ctx.keystore.save()
        if delete_whitelist:
            self.ctx.whitelist.save()

        echo_success()

    def _handle_reveal_wallet_mnemonics(self):
        wallet_name = self.__select_wallet_or_false("Wallet to reveal")
        if wallet_name == False:
            return

        questions = [
            inquirer.Password("keystore_password",
                              message='Keystore password'),
        ]
        keystore_password = self._prompt(questions)["keystore_password"]

        record = self.ctx.keystore.get_record_by_name(
            wallet_name, raise_none=True)
        mnemonics = self.ctx.keystore.get_secret(record, keystore_password)
        echo_success(mnemonics)

    def _handle_import_from_mnemonics(self):
        questions = [
            inquirer.Text(
                "mnemonics", message="Mnemonic words (splited by space)"),
            inquirer.List(
                "version", message='Wallet version', choices=[e.value for e in WalletVersionEnum],
                carousel=True, default=self.ctx.config.wton.default_wallet_version),
            inquirer.Text(
                "workchain", message='Workchain', default="0"),
            inquirer.Text("name", message='Wallet name'),
        ]
        ans = self._prompt(questions)
        mnemonics = ans["mnemonics"].split(" ")
        version = WalletVersionEnum(ans["version"])
        workchain = int(ans["workchain"])
        name = ans["name"]

        mnemonics, pub_k, _priv_k, wallet = Wallets.from_mnemonics(
            mnemonics, version, workchain)

        questions = [
            inquirer.Text(
                "comment", message='Wallet description (leave blank to skip)'),
            inquirer.Text(
                "contact_name", message='Enter name to save to whitelist (blank to skip)', default=name),
        ]
        ans = self._prompt(questions)
        comment = ans["comment"]
        contact_name = ans["contact_name"]

        if contact_name:
            contact = self.ctx.whitelist.get_contact(contact_name)
            if contact is not None:
                raise Exception(
                    f"Contact with the name '{contact_name}' already exists")

        self.ctx.keystore.add_record(name, wallet.address,
                                     pub_k.hex(), mnemonics, version,
                                     workchain, comment, save=True)

        if contact_name:
            self.ctx.whitelist.add_contact(
                contact_name, wallet.address.to_string(True), save=True)

        echo_success()

    def _handle_wallet_to_addr_pk(self):
        wallet_name = self.__select_wallet_or_false("Wallet to use")
        if wallet_name == False:
            return
        questions = [
            inquirer.Text("destination_dir",
                          message='Directory path to export into'),
            inquirer.Password("keystore_password",
                              message='Keystore password'),
        ]
        ans = self._prompt(questions)
        destination_dir = ans["destination_dir"]
        keystore_password = ans["keystore_password"]

        record = self.ctx.keystore.get_record_by_name(
            wallet_name, raise_none=True)
        mnemonics = self.ctx.keystore.get_secret(
            record, keystore_password).split(" ")
        addr, pk = Wallets.to_addr_pk(mnemonics, record.version,
                                      record.workchain)
        addr_path = os.path.join(
            destination_dir, record.name + ".addr")
        pk_path = os.path.join(destination_dir, record.name + ".pk")
        storage.save_bytes(addr_path, addr)
        storage.save_bytes(pk_path, pk)
        echo_success()

    def _handle_backup_keystore(self):
        questions = [
            inquirer.Text("backup_file_path", message='Backup filepath'),
            inquirer.Password("keystore_password",
                              message='Keystore password'),
            inquirer.Confirm(
                "is_sure",
                message='Backup stores keys in UNENCRYPTED FORM. Are you sure want to export unencrypted keys to disk?',
                default=False),

        ]
        ans = self._prompt(questions)
        backup_file_path = ans["backup_file_path"]
        keystore_password = ans["keystore_password"]
        is_sure = ans["is_sure"]

        if not is_sure:
            echo_success("Action canceled.", True)
            return

        self.ctx.keystore.backup(
            backup_file_path, keystore_password)

        echo_success()

    def __select_wallet_or_false(self, message):
        if self.ctx.keystore.records:
            questions = [
                inquirer.List(
                    "wallet",
                    message=message,
                    choices=[record.name for record in self.ctx.keystore.records],
                    carousel=True
                )
            ]
            return self._prompt(questions)["wallet"]

        echo_success("You do not have any wallets yet.")
        return False

    def __select_contact_or_false(self, message):
        if self.ctx.whitelist.contacts:
            questions = [
                inquirer.List(
                    "contact",
                    message=message,
                    choices=[
                        record.name for record in self.ctx.whitelist.contacts],
                    carousel=True
                )
            ]
            return self._prompt(questions)["contact"]

        echo_success("You do not have any contacts yet.")
        return False



from ._client import DAppTonClient, TonClient
from ._exceptions import ton_exceptions_handler, TonError

all = [
    'TonClient',
    'DAppTonClient',

    'TonError',
    'ton_exceptions_handler',
]



from collections import OrderedDict
import inquirer

from wton.tonclient.utils import KeyStores
from ._base import BaseSet
from ._keystore import KeystoreSet
from ..._utils import SharedObject, new_keystore_password_is_valid
from .._utils import echo_success, echo_error


class KeystoresSet(BaseSet):
    def __init__(self, ctx: SharedObject) -> None:
        super().__init__(ctx)
        self.ctx.keystores = KeyStores(
            self.ctx.config.wton.keystores_path)

    def _handlers(self) -> OrderedDict:
        ord_dict = OrderedDict()
        ord_dict["Open keystore"] = self._handle_open_keystore
        ord_dict["Create keystore"] = self._handle_create_keystore
        ord_dict["Restore keystore"] = self._handle_restore_keystore
        ord_dict["Back"] = self._handle_exit
        return ord_dict

    def _handle_open_keystore(self):
        if self.ctx.keystores.keystore_paths:
            keystore_name = self._prompt([
                inquirer.List("keystore_name", message="Choose keystore to use",
                              choices=self.ctx.keystores.keystore_paths, carousel=True),
            ])["keystore_name"]
            KeystoreSet(self.ctx, keystore_name).show()
        else:
            echo_success("You do not have any keystores yet.")

    def _handle_create_keystore(self):
        questions = [
            inquirer.Text(
                "name", message='Enter the name'),
            inquirer.Password(
                "password1", message='Enter the password (at least 6 symbols)'),
            inquirer.Password(
                "password2", message='Re-enter the password'),
        ]
        ans = self._prompt(questions)
        name = ans["name"]
        pass1 = ans["password1"]
        pass2 = ans["password2"]

        if self.__validate_passwords(pass1, pass2):
            self.ctx.keystores.create_new_keystore(name, pass1, save=True)
            echo_success()

    def _handle_restore_keystore(self):
        questions = [
            inquirer.List("restore_from", message='Restore from', choices=["wton backup", "ton-cli backup"]),
            inquirer.Text(
                "backup_file_path", message='Enter the path to backup file'),
            inquirer.Text(
                "name", message="Enter new name"),
            inquirer.Password(
                "password1", message='Enter the password (at least 6 symbols)'),
            inquirer.Password(
                "password2", message='Re-enter the password'),
        ]
        ans = self._prompt(questions)
        name = ans["name"]
        restore_from = ans["restore_from"]
        backup_file_path = ans["backup_file_path"]
        pass1 = ans["password1"]
        pass2 = ans["password2"]

        if self.__validate_passwords(pass1, pass2):
            if restore_from == "ton-cli backup":
                self.ctx.keystores.restore_ton_cli_keystore(
                    name, backup_file_path, pass1)
            else:
                self.ctx.keystores.restore_wton_keystore(
                    name, backup_file_path, pass1)
            echo_success()

    def __validate_passwords(self, pass1, pass2):
        if pass1 != pass2:
            echo_error("Passwords do not match.")
            return False

        if not new_keystore_password_is_valid(pass1):
            echo_error("Password must be at least 6 symbols long.")
            return False

        return True



from inquirer.render.console import ConsoleRender
from inquirer import events, errors
from inquirer.render.console import List
from .._exceptions import EscButtonPressed


ESC_BUTTONS = {"\x1b", "\x1b\x1b"}

class ModifiedConsoleRender(ConsoleRender):
    def render(self, question, answers=None):
        question.answers = answers or {}

        if question.ignore:
            return question.default

        clazz = self.render_factory(question.kind)
        render = clazz(question, terminal=self.terminal,
                       theme=self._theme, show_default=question.show_default)

        self.clear_eos()

        is_success = True

        try:
            ans = self._event_loop(render)
            return ans

        except EscButtonPressed:
            raise 

        except Exception:
            is_success = False

        finally:
            if (type(render) is List):
                self._position = len(list(render.get_options())) + 1
                self._relocate()
                self.clear_eos()
                self._print_header(render, self._get_msg_mark(is_success))
            else:
                print("")

    def _print_header(self, render, specific_msg_mark=None):
        base = render.get_header()

        header = base[: self.width - 9] + \
            "..." if len(base) > self.width - 6 else base
        default_value = " ({color}{default}{normal})".format(
            default=render.question.default, color=self._theme.Question.default_color, normal=self.terminal.normal
        )
        show_default = render.question.default and render.show_default
        header += default_value if show_default else ""

        msg_mark = "?" if specific_msg_mark is None else specific_msg_mark
        msg_template = (
            "{t.move_up}{t.clear_eol}{tq.brackets_color}[" "{tq.mark_color}" +
            msg_mark + "{tq.brackets_color}]{t.normal} {msg}"
        )

        # ensure any user input with { or } will not cause a formatting error
        escaped_current_value = str(render.get_current_value()).replace(
            "{", "{{").replace("}", "}}")
        self.print_str(
            f"\n{msg_template}: {escaped_current_value}",
            msg=header,
            lf=not render.title_inline,
            tq=self._theme.Question,
        )

    def _process_input(self, render):
        try:
            ev = self._event_gen.next()
            if isinstance(ev, events.KeyPressed):
                if ev.value in ESC_BUTTONS:
                    raise EscButtonPressed()

                render.process_input(ev.value)
        except errors.ValidationError as e:
            self._previous_error = e.value
        except errors.EndOfInput as e:
            try:
                render.question.validate(e.selection)
                raise
            except errors.ValidationError as e:
                self._previous_error = render.handle_validation_error(e)

    def _get_msg_mark(self, is_success):
        return "✓" if is_success else "✕"



import os
from typing import List

from wton.utils import storage

# workdir
DEFAULT_LOCAL_WORKDIR: str = storage.local_workdir(".wton")
GLOBAL_WORKDIR: str = storage.global_workdir()
LOCAL_WORKDIR: str = storage.find_local_workdir(".wton")
CURRENT_WORKDIR = LOCAL_WORKDIR if LOCAL_WORKDIR is not None else GLOBAL_WORKDIR

# config paths
INIT_LOCAL_CONFIG_PATH: str = os.path.join(
    DEFAULT_LOCAL_WORKDIR, "config.yaml")
GLOBAL_CONFIG_PATH: str = os.path.join(
    GLOBAL_WORKDIR, "config.yaml")
LOCAL_CONFIG_PATH: str = os.path.join(
    DEFAULT_LOCAL_WORKDIR if LOCAL_WORKDIR is None else LOCAL_WORKDIR, "config.yaml")
EXTRA_CONFIG_PATH: str = os.environ.get('WTON_CONFIG_PATH')
CURRENT_CONFIG_PATH = os.path.join(CURRENT_WORKDIR, "config.yaml")
CONFIG_PATHS: List[str] = [
    GLOBAL_CONFIG_PATH, LOCAL_CONFIG_PATH]
if EXTRA_CONFIG_PATH:
    CONFIG_PATHS.append(EXTRA_CONFIG_PATH)

# dapp
DAPP_MAINNET_GRAPHQL_URL = "https://dapp-01.tontech.io/graphql"
DAPP_MAINNET_BROADCAST_URL = "https://dapp-01.tontech.io/broadcast"
DAPP_TESTNET_GRAPHQL_URL = "https://dapp-test.tontech.io/graphql"
DAPP_TESTNET_BROADCAST_URL = "https://dapp-test.tontech.io/broadcast"

# wton
KEYSTORE_PASSWORD = os.environ.get('WTON_KEYSTORE_PASSWORD', '')



import os
from pydantic import ValidationError
import click

from ._base_cmd import cli, SharedObject
from .._utils import CustomClickException, click_echo_success
from wton import settings
from wton.config import update_config_field, unset_config_field, get_config, get_configs_with_origins, \
    Config, ConfigLocation, TonNetworkEnum
from wton.utils.exceptions import StorageError


def __list_configs(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    configs, origins = get_configs_with_origins(ctx.obj.specific_config_path)
    if configs:
        for config, origin in zip(configs, origins):
            for key, val in config.key_value(exclude_unset=True):
                click.echo(f"{origin}  {key}={val}")

    else:
        click.echo("No configs found.")

    ctx.exit()


def __current_setup(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    for key, val in ctx.obj.config.key_value():
        click.echo(f"{key}={val}")

    ctx.exit()


@cli.command()
@click.option('--list', is_flag=True, callback=__list_configs,
              expose_value=False, is_eager=True, help='Show configs values.')
@click.option('--current-setup', is_flag=True, callback=__current_setup,
              expose_value=False, is_eager=True, help='Show current configuration.')
@click.option('--global', '-g', 'config_location', flag_value=ConfigLocation.global_location.value, help='Work with a global config.')
@click.option('--local', '-l', 'config_location', default=True, flag_value=ConfigLocation.local_location.value, help='Work with the config in the current workdir. [default]')
@click.option('--file', '-f', 'config_location', help='Specific config path.', metavar='PATH')
@click.option('--network', help='Setup providers to a provided network.', type=click.Choice(TonNetworkEnum))
@click.option('--unset', is_flag=True, default=False, help='Removes [NAME] from config.')
@click.argument('name', required=False)
@click.argument('value', required=False)
@click.pass_obj
def config(shared_object: SharedObject, config_location: str, network: TonNetworkEnum, unset: bool, name: str, value: str):
    """
    Control config parameters. (check README.md for all fields info)
    """
    if config_location is None:
        config_location = settings.CURRENT_CONFIG_PATH

    if network:
        __set_network(shared_object.config, config_location, network)
        return
    elif name is None:
        raise click.MissingParameter(name)

    if not unset and not value:
        __show_value(shared_object.config, config_location, name)

    else:
        if unset:
            unset_config_field(config_location, name)
        else:
            try:
                update_config_field(config_location, name, value)
            except StorageError as e:
                raise CustomClickException(e)
            except ValidationError as e:
                msg = f"'{e.errors()[0]['loc'][0]}' {e.errors()[0]['msg']} "
                raise CustomClickException(msg)
            except ValueError as e:
                raise CustomClickException(
                    "Incorrect field name. Choices are: {}".format(", ".join(Config.field_names())))


def __show_value(config: Config, config_location: str, name: str):
    try:
        config_to_show = get_config(
            config_location) if config_location else config
    except StorageError as e:
        raise CustomClickException(e)

    if config_to_show:
        field_val = config_to_show.get_nondefault_value(name)
        if field_val:
            click.echo(field_val)


def __set_network(config: Config, config_location: str, network: TonNetworkEnum):
    update_config_field(
            config_location, "provider.dapp.network", network)



from ._mnemonic import mnemonic_new, mnemonic_to_wallet_key, mnemonic_is_valid
from ._keystore import generate_new_keystore, generate_keystore_key


__all__ = [
    'mnemonic_new',
    'mnemonic_to_wallet_key',
    'mnemonic_is_valid',

    'generate_new_keystore',
    'generate_keystore_key',

    'generate_key_pair',
]



import base64
from ...boc import Cell
from .._exceptions import ResponseError


def parse_object(obj):
    type_name = obj['@type']

    if type_name in ["tvm.list", "tvm.tuple"]:
        return [parse_object(o) for o in obj["elements"]]  # ?
    elif type_name == "tvm.stackEntryTuple":
        return parse_object(obj["tuple"])
    elif type_name == "tvm.stackEntryNumber":
        return parse_object(obj["number"])
    elif type_name == "tvm.numberDecimal":
        raise Exception("Implement me pls")
    else:
        raise Exception(f"Unknown type: {type_name}")


def parse_response_stack(pair):
    type_name = pair[0]
    val = pair[1]

    if type_name == "num":
        return int(val.replace('/0x/', ''), 16)
    elif type_name in ["list", "tuple"]:
        return parse_object(val)
    elif type_name == "cell":
        content_bytes = base64.b64decode(val)
        return Cell.one_from_boc(content_bytes)
    else:
        raise Exception(f"Unknown type: {type_name}")


def parse_response(response):
    if response["exit_code"] not in [0, 1]:
        raise ResponseError("Error response", response["exit_code"])

    arr = [parse_response_stack(pair) for pair in response["stack"]]

    return arr[0] if len(arr) == 1 else arr



from collections import OrderedDict

from ._base import BaseSet
from ._keystores import KeystoresSet
from ._whitelist import WhitelistSet


class EntrypointSet(BaseSet):
    def _handlers(self) -> OrderedDict:
        ord_dict = OrderedDict()
        ord_dict["Keystores"] = self._handle_keystores
        ord_dict["Whitelist"] = self._handle_whitelist
        ord_dict["Exit"] = self._handle_exit
        return ord_dict

    def _handle_keystores(self):
        KeystoresSet(self.ctx).show()

    def _handle_whitelist(self):
        WhitelistSet(self.ctx).show()



import json
import codecs
from typing import Any, Dict, List, Optional
import asyncio
import aiohttp

from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport
from graphql import DocumentNode
from pydantic import BaseModel


class ErrorMsg(BaseModel):
    message: str
    code: int


class DAppWrongResult(Exception):
    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return ". ".join([f"{error.message}, Code: {error.code}" for
                          error in self.errors])


class BroadcastQuery(BaseModel):
    boc: str
    timeout: int


class DAppClient:
    def __init__(self, graphql_url: str, broadcast_url: str, api_key: str):
        self.api_key = api_key
        self.broadcast_url = broadcast_url
        self.transport = AIOHTTPTransport(
            url=graphql_url, headers=self.__headers(is_json=False))

    async def query(self, queries: List[DocumentNode]) -> List[Any]:
        results = []

        async with Client(
            transport=self.transport,
            fetch_schema_from_transport=False,
        ) as session:
            for query in queries:
                result = await session.execute(query)
                results.append(result)

        return results

    async def broadcast(self, queries: List[BroadcastQuery], timeout=31, ignore_errors=False) -> List[Any]:
        results = []
        timeout = aiohttp.ClientTimeout(total=timeout)

        async with aiohttp.ClientSession(timeout=timeout) as session:
            for query in queries:
                async with session.post(self.broadcast_url, json=query.dict(), headers=self.__headers(is_json=True)) as resp:
                    results.append(await self.__parse_broadcast_response(resp, ignore_errors))

        return results

    def __headers(self, is_json):
        headers = {}

        if is_json:
            headers = {
                'Content-Type': 'application/json',
            }

        if self.api_key:
            headers['API-KEY'] = self.api_key

        return headers

    async def __parse_broadcast_response(self, resp, ignore_errors):
        try:
            resp = await resp.json()
        except Exception:  # TODO: catch correct exceptions
            if not ignore_errors:
                raise DAppWrongResult(
                    [ErrorMsg(message=resp.reason, code=resp.status)])

            return

        if "errors" in resp and resp['errors']:
            if len(resp['errors']) == 1 and 'message' in resp['errors'][0] \
                    and resp['errors'][0]['message'] == 'timeout':
                # transaction may have been sent and may be commited later
                resp['data']['status'] = 0
                return resp['data']

            else:
                if not ignore_errors:
                    errors = [ErrorMsg.parse_obj(error)
                              for error in resp['errors']]
                    raise DAppWrongResult(errors)

                return

        return resp['data']




from typing import List
from gql import gql
from graphql import DocumentNode


class DAppQueries:
    @classmethod
    def accounts(cls, account_ids: List[str], ) -> DocumentNode:
        return gql(
            """
        query {
            accounts(
                filter: {
                    id: {
                        in: [%s]
                    }
                }
            ) {
                id
                address
                acc_type_name
                state_hash
    			last_paid
    			balance
                last_trans_lt
    			code
                data
            }
        }
        """ % ", ".join([f"\"{acc_id}\"" for acc_id in account_ids]))
